"""App class to easily define the structure of a typical Voilà dashboard."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from ipywidgets import widgets, Layout
from IPython.display import display, HTML
from IPython.core.display import HTML as ipyhtml   # V. I. for urlOpen to work!!!
import base64
import plotly.express as px
import plotly.graph_objects as go
import pandas as pd
from random import randrange
from datetime import datetime
import ipyvuetify as v

try:
    from . import settings
    from . import fontsettings
    from . import sidePanel
    # from . import Button
    from . import footer
    from . import snackbar
    from . import dialogMessage
    from . import dialogYesNo
    from . import dialogWait
    from . import dialogGeneric
    from . import fab
    from . import queryStrings
except:
    import settings
    import fontsettings
    import sidePanel
    # import Button
    import footer
    import snackbar
    import dialogMessage
    import dialogYesNo
    import dialogWait
    import dialogGeneric
    import fab
    import queryStrings

from vois.vuetify import Button
    
# Convert a measure in string (if int or float --> pixels units)
def measure2str(value):
    if isinstance(value, int):
        return '%dpx' % value
    elif isinstance(value, float):
        return '%fpx' % value
    else:
        return value

#####################################################################################################################################################
# A fully customizable App framework for Voilà
#####################################################################################################################################################
class app():
    """
    App class to easily define the basic structure of a typical Voilà dashboard.
    
    An app instance displays a series of ipywidgets.Output() widgets to ease the usage of messages/dialog-boxes/etc, through these methods:
    
    :func:`~app.snackbar`
    :func:`~app.dialogMessage`
    :func:`~app.dialogYesNo`
    :func:`~app.dialogGeneric`
    :func:`~app.dialogWaitOpen`
    :func:`~app.dialogWaitClose`
    :func:`~app.fab`
    :func:`~app.downloadText`
    :func:`~app.downloadBytes`
    :func:`~app.urlOpen`
    :func:`~app.urlParameter`
    :func:`~app.urlParameter`

    The main content of the dashboard should be displayed in the **app.outcontent** Output widget that completely fills the empty space of the dashboard between the title bar and the footer bar.
    
    Parameters
    ----------
    title : str, optional
        Main title of the application to be displayed on the title bar (default is 'Title of the dashboard')
    titlesvg : str, optional
        SVG string to use as application title when the title string is empty (default is '')
    titlesvgclass : str, optional
        Class margins and padding to apply to the titlesvg drawing (default is 'pa-0 ma-0')
    titlecredits : str, optional
        Credits string to be displayed on the right side of the title bar (default is '')
    titlecredits2 : str, optional
        Secondary credits string to be displayed on the right side of the title bar (default is '')
    titlestyle : str, optional
        CSS style to apply to the main title of the dashboard (default is '', an example could be: 'font-family: "Times New Roman", Times, serif; font-weight: bold; font-size: 22px;')
    titlespacestyle : str, optional
        CSS style to apply to the space at the left of the title (default is 'width: 50px; min-width: 50px;'). It can be useful to move the title more on the centre of the title bar, by providing a titlespacestyle like 'width: 200px;'
    titleheight : int, optional
        Height of the title bar in pixels (default is 70 pixels)
    totalheight : int, optional
        Total height in pixels of the page (default is 985 pixels which is coherent with a FullHD screen dimension)
    dark : bool, optional
        If True the title text color is settings.textcolor_dark, if False it is settings.textcolor_nodark (default is settings.dark_mode)
    backcolor : str, optional
        Background color of the title bar (default is settings.color_second)
    titlewidth : str, optional
        Width of the part of the title bar  that contains the main title of the dashboard (default is '600px', other values could be, for instance: '50%')
    footercolor : str, optional
        Background color to use for the footer bar displayed in the bottom part of the screen (default is 'lightgrey')
    backgroundimageurl : str, optional
        URL of the optional image to display as background of the title bar (default is None)
    backgroundimageposition : str, optional
        Defines the way the backgroundimage in the title is cropped (default is 'center center'). See `Vuetify.js v-img <https://vuetifyjs.com/en/api/v-img/#props-position>`_  and `CSS background-position documentation <https://developer.mozilla.org/en-US/docs/Web/CSS/background-position>`_ for help.
    logourl : str, optional
        URL of the image or SVG to use as a logo in the right end side of the title bar (default is European Commission logo)
    logomaxwidth : int, optional
        Maximum width in pixels of the logo image (default is 80 pixels)
    logomaxheight : int, optional
        Maximum height in pixels of the logo image (default is 60 pixels)
    titletabs : list of strings, optional
        List of tabs to be added to the title bar as the main visualization options of the dashboard (default is ['Tab 1', 'Tab 2', 'Tab 3'])
    titletabsactive : int, optional
        Index of the tab to activate at start (default is 0)
    titletabsstile : str, optional
        CSS style to apply to the tabs in the title bar (default is 'font-weight: bold;')
    titletabsactiveparameter : str, optional
        Name of the URL parameter to read for setting the activetab (default is None)
    titletabscolor : str, optional
        Color of the selected tab in the title tabs (default is settings.color_first)
    titletabsdark : bool, optional
        If True the text color of unselected tabs is settings.textcolor_dark, if False it is settings.textcolor_nodark (default is False)
    footertext : str, optional
        Text to display in the footer tab (default is '<current year> - Joint Research Centre')
    footerbuttons : list of strings, optional
        List of strings containing the caption of the buttons to display in the footer tab (default is ['Home', 'About Us', 'Team', 'Services', 'Blog', 'Contact Us'])
    footerheight : int or float or str optional
        Height of the footer bar. If an integer or a float is passed, the height is intended in pixels units, otherwise a string containing the units must be passed (example: '4vh'). Default is 68 for 68 pixels
    footercopyright : bool, optional
        If True adds the copyright symbol to the footer text (default is True)
    footerdark : bool, optional
        If True the footer text color is settings.textcolor_dark, if False it is settings.textcolor_nodark (default is settings.dark_mode)
    footercredits : str, optional
        Text for footer credits button (default is '')
    footercreditstooltip : str, optional
        Tooltip for the footer credits button (default is '')
    footercreditsurl : str, optional
        URL to open when the user clicks on the footer credits button (default is '')
    sidepaneltitle : str, optional
        Title of the left side panel (default is 'Settings')
    sidepanelwidth : int, optional
        Width in pixels of the left side panel (default is 400). If the size is 0, the icon to open the sidepanel will not be added to the title bar
    sidepaneltext : str, optional
        Text to display in the left side panel (default is '')
    sidepanelcontent : list of ipywidgets, optional
        List of ipywidgets object to display in the left side panel (default is [])
    sidepaneldark : bool, optional
        If True the title text color of the sidePanel is black, otehrwise is white (default is settings.dark_mode)
    sidepanelbackdark : bool, optional
        If True the background color of the sidePanel is black, otehrwise is white (default is settings.dark_mode)
    minipanelicons : list of strings, optional
        List of string containing the names of the icons to display in the minipanel that is displayed on the left side of the footer tab (default is [])
    minipaneltooltips : list of strings, optional
        List of tooltips to set for the icons of the minipanel (default is [])
    minipanelopen : bool, optional
        If True the minipanel is dopened on startup of the app (default is False)
    minipanellarge : bool, optional
        If True, the minipanel icons are displayed in the large version (default is True)
    minipanelbuttoncolor : str, optional
         Color of the 'three vertical points icon' that opens/closes the minipanel (default is the textcolor_notdark defined in the settings.py module)
    minipaneliconscolor : str, optional
         Color of the icons in the minipanel (default is the textcolor_notdark defined in the settings.py module)
    onclicktab : function, optional
        Python function to call when the user clicks on one of the tabs of the title bar. The function will receive a parameter of type string containing the text of the tab
    onclickcredits : function, optional
        Python function to call when the user clicks on the credits button on the title bar. The function will receive no parameters
    onclickcredits2 : function, optional
        Python function to call when the user clicks on the secondary credits button on the title bar. The function will receive no parameters
    onclicklogo : function, optional
        Python function to call when the user clicks on the logo image on the title bar. The function will receive no parameters
    onclickfooter : function, optional
        Python function to call when the user clicks on one of the buttons of the footer bar. The function will receive a parameter of type string containing the text of the button
    onclickminipanel : function, optional
        Python function to call when the user clicks on one of the icons of the minipanel in the footer bar. The function will receive a parameter of type int containing the index of the icon
    fullscreen : bool, optional
        If True the app will be displayed in fullscreen mode (default is False). In fullscreen mode the app will occupy all the available space on the web page (the titlebar will be aligned on top, the footer bar will be aligned on the bottom of the page, and the outcontent will occupy all the intermediate space between the title bar and the footer bar, irrescpective of the value passed in the totalheight parameter), and the positioning of the elements will be fully responsive.
        
        
    Attributes
    ----------
    outcontent : ipywidgets.Output instance
        This is the output widget where the content of the application can be displayed

        To visually highlight the app.outcontent Output widget, this line of code can be executed::

            g_app.outcontent.layout.border = '1px solid lightgrey'

        After the execution of that line, the border of the g_app.outcontent will be visible. To reset the border to the empty line, this line of code can be executed::

            g_app.outcontent.layout.border = ''


    Example
    -------
    Creation of an app class to define the structure of a Voilà dashboard::
      
        from vois.vuetify import app, settings
        import ipyvuetify as v
        
        # Change global settings
        settings.dark_mode      = False
        settings.color_second   = '#68aad2'
        settings.color_first    = '#1c4056'
        settings.button_rounded = False

        # Click on a tab of the title
        def on_click_tab(arg):
            g_app.snackbar(arg)

        # Click on the credits text
        def on_click_credits():
            g_app.snackbar('CREDITS')

        # Click on the logo
        def on_click_logo():
            g_app.snackbar('LOGO')

        # Click on the footer buttons
        def on_click_footer(arg):
            g_app.snackbar(arg)

        # Click on the footer minipanel
        def on_click_minipanel(index):
            g_app.snackbar(str(index))


        g_app = app.app(title='Energy consumption example dashboard',
                        titlecredits='Created by Unit I.3',
                        titlewidth='60%',
                        footercolor='#1c4056',
                        titletabs=['Chart', 'Table', 'Static Map', 'Dynamic Map'],
                        titletabscolor='#60b3e8',
                        dark=True,
                        footerdark=True,
                        footercredits='Data credits',
                        footercreditstooltip='Eurostat - European Commission',
                        footercreditsurl='https://ec.europa.eu/eurostat/data/database',
                        backcolor='#1c4056',
                        sidepaneltitle='Help',
                        sidepaneltext='Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.',
                        sidepanelcontent=[v.Icon(class_='pa-0 ma-0 ml-2', children=['mdi-help'])],
                        sidepaneldark=True,
                        sidepanelbackdark=False,
                        minipanelicons=['mdi-chart-bar', 'mdi-table-large',
                                        'mdi-map-legend', 'mdi-file-chart-outline'],
                        minipaneltooltips=['Download Chart', 'Download Table',
                                           'Download Map', 'Generate Report in Word'],
                        minipanelbuttoncolor='white',
                        onclickminipanel=on_click_minipanel,
                        onclicktab=on_click_tab,
                        onclickcredits=on_click_credits,
                        onclicklogo=on_click_logo,
                        onclickfooter=on_click_footer)

        g_app.show()
        
    .. figure:: figures/app.png
       :scale: 100 %
       :alt: app widget

       App structure example

   """
        
    def __init__(self, 
                 title='Title of the dashboard',
                 titlesvg='',
                 titlesvgclass='pa-0 ma-0',
                 titlecredits='',
                 titlecredits2='',
                 titlestyle='',      # 'font-family: "Times New Roman", Times, serif; font-weight: bold; font-size: 22px;'
                 titlespacestyle='width: 50px; min-width: 50px;',
                 titleheight=70,
                 totalheight=985,
                 dark=settings.dark_mode,
                 backcolor=settings.color_second,
                 titlewidth= '600px',    # '50%'
                 footercolor='lightgrey',
                 backgroundimageurl=None,     #'https://picsum.photos/id/1050/1920/1080?grayscale'
                 backgroundimageposition='center center',
                 logourl='https://jeodpp.jrc.ec.europa.eu/services/shared/Notebooks/images/European_Commission.svg',
                 logomaxwidth  = 80,
                 logomaxheight = 60,

                 titletabs=['Tab 1', 'Tab 2', 'Tab 3'],
                 titletabsdark=False,
                 titletabsactive=0,                       # Index of the tab to activate at start
                 titletabsstile='font-weight: bold;',     #'font-family: "Times New Roman", Times, serif; font-weight: bold;'
                 titletabsactiveparameter=None,           # Name of the URL parameter to read for setting the activetab 
                 titletabscolor=settings.color_first,     # Color of the active tab in the title tabs
                 

                 footertext='%d - Joint Research Centre'%(datetime.now().year),
                 footerbuttons=['Home', 'About Us', 'Team', 'Services', 'Blog', 'Contact Us'],
                 footerheight=68,
                 footercopyright=True,
                 footerdark=False,
                 footercredits='',           # Text for footer credits button
                 footercreditstooltip='',    # Tooltip for the footer credits button
                 footercreditsurl='',        # URL to open when the user click on the footer credits button

                 sidepaneltitle='Settings',
                 sidepanelwidth=400,
                 sidepaneltext='',
                 sidepanelcontent=[],
                 sidepaneldark=settings.dark_mode,
                 sidepanelbackdark=settings.dark_mode,
                 
                 minipanelicons=[],          # See footer.py
                 minipaneltooltips=[],
                 minipanelopen=False,
                 minipanellarge=True,
                 minipanelbuttoncolor=settings.textcolor_notdark,
                 minipaneliconscolor=settings.textcolor_notdark,

                 onclicktab=None,            # Function with 1 string argument string containing the text of the tab
                 onclickcredits=None,        # Function with 0 arguments
                 onclickcredits2=None,       # Function with 0 arguments
                 onclicklogo=None,           # Function with 0 arguments
                 onclickfooter=None,         # Function with 1 string argument string containing the text of the button
                 onclickminipanel=None,      # Function with 1 string argument integer containing the index of the icon
                 fullscreen=False):          # If True the App will fill all the page!
        
        # Storing input parameters
        self.title = title
        self.titlecredits = titlecredits
        self.titlecredits2 = titlecredits2
        self.titlestyle = titlestyle
        self.titleheight = measure2str(titleheight)
        self.totalheight = measure2str(totalheight)
        self.dark = dark
        self.backcolor = backcolor
        self.titlewidth = measure2str(titlewidth)
        self.footercolor = footercolor
        self.backgroundimageurl = backgroundimageurl
        self.backgroundimageposition = backgroundimageposition
        self.logourl = logourl
        self.logomaxwidth = logomaxwidth
        self.logomaxheight = logomaxheight

        self.titletabsstile = titletabsstile
        self.titletabs = titletabs
        self.titletabsactive = titletabsactive
        self.titletabsactiveparameter = titletabsactiveparameter
        self.titletabscolor = titletabscolor

        self.footertext = footertext
        self.footerbuttons = footerbuttons
        self.footerheight = measure2str(footerheight)
        self.footercopyright = footercopyright
        self.footerdark = footerdark
        self.footercredits=footercredits
        self.footercreditstooltip=footercreditstooltip
        self.footercreditsurl=footercreditsurl

        self.sidepaneltitle = sidepaneltitle
        self.sidepanelwidth = sidepanelwidth
        self.sidepaneltext = sidepaneltext
        self.sidepanelcontent = sidepanelcontent
        self.sidepaneldark = sidepaneldark
        self.sidepanelbackdark = sidepanelbackdark

        self.onclicktab = onclicktab
        self.onclickcredits = onclickcredits
        self.onclickcredits2 = onclickcredits2
        self.onclicklogo = onclicklogo
        self.onclickfooter = onclickfooter
        
        self.fullscreen = fullscreen
        
        self.content_vbackground = None     # v.Html object to display an image in the background
        self.content_panels_dict = {}       # Dict to retrieve panels by name
        self.content_panels      = []       # List of panels to overlay to the background
        
        
        if self.dark:
            buttontext = settings.textcolor_dark
            styletext = 'color: ' + settings.textcolor_dark + ';'
        else:
            buttontext = settings.textcolor_notdark
            styletext = 'color: ' + settings.textcolor_notdark + ';'
        
        self.contentheight = 'calc(calc(%s - %s) - %s)' % (self.totalheight, self.titleheight, self.footerheight)

        # Creation of Output widgets
        self.outservice   = widgets.Output(layout=Layout(width='0px', height='0px'))  # Available for the user
        self.outdialogs   = widgets.Output(layout=Layout(width='0px', height='0px'))  # Used for snackbars, fab and dialogs
        self.outdownload  = widgets.Output(layout=Layout(width='0px', height='0px'))  # Used for downloads
        self.outfabs      = widgets.Output(layout=Layout(width='0px', height='0px'))  # Used for fabs
        self.outurlopen   = widgets.Output(layout=Layout(width='0px', height='0px'))  # Used for urlOpen
        self.outurlupdate = widgets.Output(layout=Layout(width='0px', height='0px'))  # Used for urlUpdate
        self.outplotly    = widgets.Output(layout=Layout(width='0px', height='0px'))  # Used for initializing plotly
        self.outpanel     = widgets.Output(layout=Layout(width='0px', height='0px'))  # Used for the side panel
        self.outpanels    = widgets.Output(layout=Layout(width='0px', height='0px'))  # used for the panels 
        
        if self.fullscreen:
            self.outcontent  = widgets.Output(layout=Layout(width='100vw', height='calc(calc(100vh - %s) - %s)' % ( self.titleheight, self.footerheight) ) )
        else:
            self.outcontent  = widgets.Output(layout=Layout(width='99.7%', height=self.contentheight))

        self.outfooter = widgets.Output(layout=Layout(width='99.7%', height='calc(%s + 6px)' % self.footerheight))

        # One instance of the dialogWait
        self.waitdlg = None
        
        # Read the parameters passed on the URL 
        self.parameters = queryStrings.readParameters()
        if not self.titletabsactiveparameter is None:
            self.titletabsactive = int(self.urlParameter(self.titletabsactiveparameter, self.titletabsactive))
                
        
        # Side panel
        self.panel = sidePanel.sidePanel(title=self.sidepaneltitle, text=self.sidepaneltext, content=self.sidepanelcontent,
                                         dark=self.sidepaneldark, backdark=self.sidepanelbackdark,
                                         width=self.sidepanelwidth, right=False, output=self.outpanel, zindex=9999, onclose=self.__internal_panelclosed)

        # App bar containing NavIcon, Title, Tabs, Credits and Logo
        if self.sidepanelwidth > 0:
            self.abi = v.AppBarNavIcon(style_=styletext)
            self.abi.on_event('click', self.__internal_onclickNavIcon)
            self.abih = v.Html(tag='div', class_="pa-0 ma-0 ml-2", children=[self.abi], style_='overflow: hidden; %s' % titlespacestyle)
        else:
            self.abih = v.Html(tag='div', class_="pa-0 ma-0", children=[' '], style_='overflow: hidden; %s' % titlespacestyle)

        if len(titlesvg) > 0:
            outsvg = widgets.Output()
            with outsvg:
                display(HTML(titlesvg))
            self.abt = v.ToolbarTitle(children=[outsvg], class_=titlesvgclass, style_='%s %s' % (styletext, self.titlestyle) )
        else:
            self.abt = v.ToolbarTitle(children=[self.title], style_='%s %s' % (styletext, self.titlestyle) )
            
        
        if self.onclickcredits is None:
            self.credits = v.Btn(text=True, color=buttontext, dark=self.dark, children=[self.titlecredits], rounded=settings.button_rounded, style_='text-transform: none; cursor: initial;')
        else:
            self.credits = v.Btn(text=True, color=buttontext, dark=self.dark, children=[self.titlecredits], rounded=settings.button_rounded, style_='text-transform: none;')
            self.credits.on_event('click', self.__internal_onclickCredits)

        self.credits2 = v.Html(tag='div', children=[''])
        if len(self.titlecredits2) > 0:
            if self.onclickcredits2 is None:
                self.credits2 = v.Btn(text=True, color=buttontext, dark=self.dark, children=[self.titlecredits2], rounded=settings.button_rounded, style_='text-transform: none; cursor: initial;')
            else:
                self.credits2 = v.Btn(text=True, color=buttontext, dark=self.dark, children=[self.titlecredits2], rounded=settings.button_rounded, style_='text-transform: none;')
                self.credits2.on_event('click', self.__internal_onclickCredits2)
            
        self.tlist = []
        for t in self.titletabs:
            telem = v.Tab(style_='%s %s' % ('', self.titletabsstile), children=[t])
            telem.on_event('click', self.__internal_onclickTab)
            self.tlist.append(telem)
        self.tabs = v.Tabs(v_model=self.titletabsactive, dark=titletabsdark, color=self.titletabscolor, background_color='transparent', 
                           align_with_title=True, children=self.tlist, height=self.titleheight, min_height=self.titleheight, max_height=self.titleheight)

        s = v.Spacer()
        logostyle = ''
        if not self.onclicklogo is None: logostyle = 'cursor: pointer;'
        self.logoimg = v.Img(src=self.logourl, class_='pa-0 ma-0', style_=logostyle, contain=True, height=self.logomaxheight, width=self.logomaxwidth, max_height=self.logomaxheight, max_width=self.logomaxwidth)
        self.logoimg.on_event('click', self.__internal_onclickLogo)
        
        card1 = v.Card(height=titleheight, color='transparent', elevation=0, children=[self.abih],    class_="d-flex align-center")
        card2 = v.Card(height=titleheight, color='transparent', elevation=0, children=[self.abt],     class_="d-flex align-center", style_='overflow: hidden; width: %s;' % self.titlewidth)
        card3 = v.Card(height=titleheight, color='transparent', elevation=0, children=[self.tabs],    class_="d-flex align-center")
        card4 = v.Card(height=titleheight, color='transparent', elevation=0, children=[self.credits, self.credits2], class_="d-flex align-center")
        card5 = v.Card(height=titleheight, color='transparent', elevation=0, children=[self.logoimg], class_="d-flex align-center mr-1")
        
        if len(self.titletabs) > 0:
            r = v.Row(justify="space-between", children=[card1,card2,card3,s,card4,card5], class_="pa-0 ma-0")
        else:
            r = v.Row(justify="space-between", no_gutters=True, children=[card1,card2,s,card4,card5], class_="pa-0 ma-0")
        if (not self.backgroundimageurl is None) and (len(self.backgroundimageurl) > 0):
            self.appbar = v.Img(src=self.backgroundimageurl, position=self.backgroundimageposition, children=[r], class_='pa-0 ma-0', style_='height: %s; overflow: hidden;' % self.titleheight, max_height=self.titleheight)
        else:
            self.appbar = v.Footer(color=backcolor, padless=True, children=[r], class_='pa-0 ma-0', style_='height: %s; overflow: hidden;' % self.titleheight, max_height=self.titleheight)

    
        # Footer
        if len(self.footerbuttons) > 0:
            self.f = footer.footer(text=self.footertext, color=self.footercolor, copyright=self.footercopyright, dark=self.footerdark,
                                   minipanelicons=minipanelicons, minipaneltooltips=minipaneltooltips, minipanelopen=minipanelopen,
                                   minipanellarge=minipanellarge, minipanelbuttoncolor=minipanelbuttoncolor, 
                                   minipaneliconscolor=minipaneliconscolor, onclickminipanel=onclickminipanel,
                                   footercredits=self.footercredits, footercreditstooltip=self.footercreditstooltip, onclickcredits=self.__internal_onclickFooterCredits,
                                   buttons=self.footerbuttons, marginy=2, height=self.footerheight, onclick=self.onclickfooter) #, output=self.outfooter)
        else:
            self.f = footer.footer(text=self.footertext, color=self.footercolor, copyright=self.footercopyright, dark=self.footerdark,
                                   minipanelicons=minipanelicons, minipaneltooltips=minipaneltooltips, minipanelopen=minipanelopen,
                                   minipanellarge=minipanellarge, minipanelbuttoncolor=minipanelbuttoncolor, 
                                   minipaneliconscolor=minipaneliconscolor, onclickminipanel=onclickminipanel,
                                   footercredits=self.footercredits, footercreditstooltip=self.footercreditstooltip, onclickcredits=self.__internal_onclickFooterCredits,
                                   marginy=2, height=self.footerheight, onclick=self.onclickfooter) #, output=self.outfooter)


        # Initialize Plotly with a 0x0 chart!
        with self.outplotly:
            data = {'name': ['mike', 'mike', 'cindy', 'cindy'],
                    'week': ['Week 1', 'Week 2', 'Week 1', 'Week 2'],
                    'perc': [0.45, 0.15, 0.25, 0.28]}
            df = pd.DataFrame(data)
            fig = px.bar(df, x="name", y="perc", color="week", title="", barmode='stack')
            fig.show()
    
    
    # Set the background image for the outcontent
    def contentBackground(self, imageUrl=None):
        """
        Sets the background image for the outcontent output widget

        Parameters
        ----------
        imageUrl : str, optional
            URL of the optional image to display as background of the outcontent output widget (default is None)
        """
        if imageUrl is None:
            self.content_vbackground = None
            self.outcontent.clear_output()
        else:
            self.content_vbackground = v.Html(tag='div', children=[''], class_="pa-0 ma-0", style_="width: 100vw; height: %s; border: 0px solid red; background-image: url('%s'); ; background-repeat: repeat;" % (self.outcontent.layout.height, imageUrl))
            self.outcontent.clear_output()
            with self.outcontent:
                display(self.content_vbackground)


    # Remove all panels
    def contentResetPanels(self):
        """
        Resets the list of panels to display on top of the outcontent output widget
        """
        self.content_panels_dict = {}
        self.content_panels      = []
        self.outpanels.clear_output()
        if (not self.content_vbackground is None) :
            self.content_vbackground.children = []

            
    # Adds a new panel
    def contentAddPanel(self, width, height, left, top, border='1px solid black', backcolor='white', name='', number=None,
                              title='', titleround=False, titlewidth='8vw', titleheight='2.4vh', titlefontsize='1.8vh', titlefontweight=500,
                              icon='', icontooltip='', iconcolor='black', icononclick=None):
        """
        Adds a new panel to overlay on top of the outcontent output widget.
        
        Parameters
        ----------
        width : str
            Width of the panel (can be in any CSS coordinates, examples: '300px', '40vw' or 'calc(100vw - 600px)')
        height : str
            Height of the panel (can be in any CSS coordinates, examples: '300px', '40vh' or 'calc(100vh - 400px)')
        left : str
            Position of the panel from the left border of the outcontent output widget (can be in any CSS coordinates, examples: '300px' or '10vw')
        top : str
            Position of the panel from the top border of the outcontent output widget (can be in any CSS coordinates, examples: '300px' or '10vh')
        border : str, optional
            Border of the panel (default is '1px solid black')
        backcolor : str, optional
            Background color of the panel (default is 'white'). Please be aware that colors different from 'white' work badly if used for panels that will contain ipywidgets or ipyvuetify widgets. Transparent colors (for example: #ffffff00) or semi-transparent colors (for example: #ffffff55) work well if the panel contains only one of the SVG charts of the vois library (svhHeatmap, svgBubblesChart, etc.)
        name : str, optional
            Name of the panel (default is '')
        number : int, optional
            Number to assign to the panel, to generate unique CSS class names (default is None which means that the number is automatically generated)
        title : str, optional
            Title to show in the top-left border of the panel (default is '')
        titleround : bool, optional
            If True the sides of the title are rounded (default is False)
        titlewidth : str, optional
            Width of the title area (default is '8vw')
        titleheight : str, optional
            Height of the title area (default is '2.4vh')
        titlefontsize : str, optional
            Size of the font to use to display the title (default is '1.8vh')
        titlefontweight : int, optional
            Weight of the font to use to display the title (default is 500)
        icon : str, optional
            Name of an icon to display in the title bar of the panel (default is None)
        icontooltip : str, optional
            Tooltip to show when hover on the icon (default is '')
        iconcolor : str, optional
            Color of the icon (default is 'black')
        icononclick : function, optional
            Python function to call when the user clicks on the icon. The function will receive as parameter the name of the panel (default is None)
        """
        out = widgets.Output(layout=Layout(width=width, height=height, border=border))
        if number is None: num = randrange(999999)
        else:              num = int(Number)
        classname = "%s_panel_out_%d" % (name,num)
        out.add_class(classname)
        out.clear_output()
        with self.outpanels:
            display(HTML('<style> .%s { background-color: %s !important; }</style>' % (classname, backcolor) ))
        panel = v.Html(tag='div', children=[out], class_="pa-0 ma-0", style_='width: %s; height: %s; border: none; position: absolute; top: %s; left: %s;' % (width, height, top, left))
        self.content_panels.append(panel)
        if len(name) > 0:
            self.content_panels_dict[name] = [panel]
        
        if len(title) > 0 :
            radius = "0px 0px 0px 0px"
            if titleround :
                radius = "4px 4px 0px 0px"
            border   = 'border-radius: %s; border: %s; border-style: solid solid none solid;' % (radius, border)
            position = 'position: absolute; top: calc(%s - %s); left: %s;' % (top, titleheight, left)
            
            children = [title]
            b = None
            if len(icon) > 0:
                b = Button('', textcolor='#ffffff00', onclick=icononclick, argument=name, width=36, small=False, disabled=False, height=titleheight, outlined=False,
                                  tooltip=icontooltip, selected=False, icon=icon, iconleft=True, iconcolor=iconcolor, rounded=False)
                children = [v.Row(no_gutters=True, justify="space-between", children=[title,b.draw()])]
            
            titlecard = v.Card(width=titlewidth, elevation=0, color=backcolor, children=children,
                               style_='font-family: %s; line-height: %s; font-size: %s; font-weight: %d; padding-left: 8px; text-align: start; %s %s' % (fontsettings.font_name, titleheight, titlefontsize, titlefontweight, border, position) )
            self.content_panels.append(titlecard)
            if len(name) > 0:
                self.content_panels_dict[name].append(titlecard)
            
        return out

    

    # Change settings of an existing panel given its name
    def contentSetPanel(self, name, width, height, left, top, border='1px solid black', backcolor='white',
                             titleround=False, titlewidth='8vw', titleheight='2.4vh', titlefontsize='1.8vh', titlefontweight=500,
                             icon='', icontooltip='', iconcolor='black', icononclick=None):
        """
        Change position, sizing and colors of a panel given its name.
        
        Parameters
        ----------
        name : str
            Name of the panel to modify
        width : str
            Width of the panel (can be in any CSS coordinates, examples: '300px', '40vw' or 'calc(100vw - 600px)')
        height : str
            Height of the panel (can be in any CSS coordinates, examples: '300px', '40vh' or 'calc(100vh - 400px)')
        left : str
            Position of the panel from the left border of the outcontent output widget (can be in any CSS coordinates, examples: '300px' or '10vw')
        top : str
            Position of the panel from the top border of the outcontent output widget (can be in any CSS coordinates, examples: '300px' or '10vh')
        border : str, optional
            Border of the panel (default is '1px solid black')
        backcolor : str, optional
            Background color of the panel (default is 'white'). Please be aware that colors different from 'white' work badly if used for panels that will contain ipywidgets or ipyvuetify widgets. Transparent colors (for example: #ffffff00) or semi-transparent colors (for example: #ffffff55) work well if the panel contains only one of the SVG charts of the vois library (svhHeatmap, svgBubblesChart, etc.)
        titleround : bool, optional
            If True the sides of the title are rounded (default is False)
        titlewidth : str, optional
            Width of the title area (default is '8vw')
        titleheight : str, optional
            Height of the title area (default is '2.4vh')
        titlefontsize : str, optional
            Size of the font to use to display the title (default is '1.8vh')
        titlefontweight : int, optional
            Weight of the font to use to display the title (default is 500)
        icon : str, optional
            Name of an icon to display in the title bar of the panel (default is None)
        icontooltip : str, optional
            Tooltip to show when hover on the icon (default is '')
        iconcolor : str, optional
            Color of the icon (default is 'black')
        icononclick : function, optional
            Python function to call when the user clicks on the icon. The function will receive as parameter the name of the panel (default is None)
        """
        if name in self.content_panels_dict:
            panellist = self.content_panels_dict[name]
            panel     = panellist[0]
            titlecard = None
            if len(panellist) > 1:
                titlecard = panellist[1]
                
            if not titlecard is None and type(titlecard.children[0]) == v.generated.Row:
                title = titlecard.children[0].children[0]
            else:
                title = titlecard.children[0]
                
            if not titlecard is None:
                if len(icon) > 0:
                    b = Button('', textcolor='#ffffff00', onclick=icononclick, argument=name, width=36, small=False, disabled=False, height=titleheight, outlined=False,
                                      tooltip=icontooltip, selected=False, icon=icon, iconleft=True, iconcolor=iconcolor, rounded=False)
                    titlecard.children = [v.Row(no_gutters=True, justify="space-between", children=[title,b.draw()])]
                else:
                    titlecard.children = [title]
            
            classname = panel.children[0]._dom_classes[0]
            with self.outpanels:
                display(HTML('<style> .%s { background-color: %s !important; }</style>' % (classname, backcolor) ))
            panel.children[0].layout = Layout(width=width, height=height, border=border)
            panel.style_ = 'width: %s; height: %s; border: none; position: absolute; top: %s; left: %s;' % (width, height, top, left)
            panel.color = backcolor
            if not titlecard is None:
                radius = "0px 0px 0px 0px"
                if titleround :
                    radius = "4px 4px 0px 0px"
                border   = 'border-radius: %s; border: %s; border-style: solid solid none solid;' % (radius, border)
                position = 'position: absolute; top: calc(%s - %s); left: %s;' % (top, titleheight, left)
                titlecard.color  = backcolor
                titlecard.width  = titlewidth
                titlecard.style_ = 'font-family: %s; line-height: %s; font-size: %s; font-weight: %d; padding-left: 8px; text-align: start; %s %s' % (fontsettings.font_name, titleheight, titlefontsize, titlefontweight, border, position)

                
    
    # Display the app
    def show(self):
        """Display the app"""
        if (not self.content_vbackground is None) and (len(self.content_panels) > 0):
            self.content_vbackground.children = self.content_panels
            
        # Fix Firefox bug with white areas: with this fix all v.Img objects MUST declare the width and height to work correctly
        with self.outdialogs:
            display(HTML('<style>.v-responsive__sizer { padding-bottom: 0px !important; }</style>'))
            
        if self.fullscreen:
            outFullscreen = widgets.Output(layout=Layout(width='100vw', height='100vh'))
            #outFullscreen.add_class('box')
            with self.outdialogs:
                display(HTML('<style>.jp-OutputPrompt { min-width: 0px; border: 0px; }</style>'))
            with outFullscreen:
                display(self.appbar)
                display(widgets.HBox([self.outservice,self.outdialogs,self.outdownload,self.outfabs,self.outurlopen,self.outurlupdate,self.outplotly,self.outpanel,self.outpanels]))
                display(self.outcontent)
                display(self.f.draw())
                
            run_in_voila = True
            try:
                from voila.utils import get_query_string
            except:
                run_in_voila = False
                    
            transition = 'dialog-fade-transition'
            persistent = False
            if run_in_voila: persistent = True

            background = v.Card(children=[outFullscreen])
            dialog = v.Dialog(v_model=True, fullscreen=True, transition=transition, persistent=persistent, no_click_animation=True, children=[background], style_='z-index:20001;')
            display(dialog)
        else:
            display(self.appbar)
            display(widgets.HBox([self.outservice,self.outdialogs,self.outdownload,self.outfabs,self.outurlopen,self.outurlupdate,self.outplotly,self.outpanel]))
            display(self.outcontent)
            #display(self.outfooter)
            display(self.f.draw())

    # Sets the active tab
    def setActiveTab(self, index):
        """Set the active tab of the title bar of the app"""
        if index >= 0 and index < len(self.tlist):
            self.titletabsactive = index
            self.tabs.v_model = self.titletabsactive
        
    # Manage click on the navigation icon to open the side panel
    def __internal_onclickNavIcon(self, *args):
        self.panel.show()
    
    # Manage click on the logo
    def __internal_onclickLogo(self, *args):
        if not self.onclicklogo is None:
            self.onclicklogo()
            
    # Manage click on the title tabs 
    def __internal_onclickTab(self, widget, event, data):
        if self.onclicktab:
            i = self.tlist.index(widget)
            self.onclicktab(self.titletabs[i])
            
    # Manage click on the title credits button
    def __internal_onclickCredits(self, widget, event, data):
        if not self.onclickcredits is None:
            self.onclickcredits()
        
    # Manage click on the title credits2 button
    def __internal_onclickCredits2(self, widget, event, data):
        if not self.onclickcredits2 is None:
            self.onclickcredits2()
            
    # When the panel has bee closed
    def __internal_panelclosed(self):
        self.outpanel.clear_output()
        with self.outpanel:
            display(self.panel.nav)
            
    # Manage click on the footer credits button
    def __internal_onclickFooterCredits(self, *args):
        if len(self.footercreditsurl) > 0:
            self.urlOpen(self.footercreditsurl, target='_blank')
        
        
    # Display something in the service Output of this application
    def display(self, arg):
        """Display something in the service Output of this application"""
        with self.outservice:
            display(arg)

            
    # Display a message in a snackbar
    def snackbar(self, *args, **kwargs):
        """Display a message in a snackbar. See :func:`~snackbar.snackbar` for the list of parameters."""
        #if not 'color' in kwargs: kwargs['show'] = 
        kwargs['show']   = True
        kwargs['output'] = self.outdialogs
        snackbar.snackbar(*args, **kwargs)
        
    # Display a dialogMessage
    def dialogMessage(self, *args, **kwargs):
        """Display a dialogMessage. See :func:`~dialogMessage.dialogMessage` for the list of parameters."""
        kwargs['show']   = True
        kwargs['output'] = self.outdialogs
        dialogMessage.dialogMessage(*args, **kwargs)
        
    # Display a dialogYesNo
    def dialogYesNo(self, *args, **kwargs):
        """Display a dialogYesNo. See :func:`~dialogYesNo.dialogYesNo` for the list of parameters."""
        kwargs['show']   = True
        kwargs['output'] = self.outdialogs
        dialogYesNo.dialogYesNo(*args, **kwargs)
        
    # Display a dialogGeneric
    def dialogGeneric(self, *args, **kwargs):
        """Display a dialogGeneric. See :func:`~dialogGeneric.dialogGeneric` for the list of parameters."""
        kwargs['show']   = True
        kwargs['output'] = self.outdialogs
        dialogGeneric.dialogGeneric(*args, **kwargs)
        
    # Open a dialogWait
    def dialogWaitOpen(self, *args, **kwargs):
        """Open a dialogWait. See :func:`~dialogWait.dialogWait` for the list of parameters."""
        kwargs['output'] = self.outdialogs
        self.waitdlg = dialogWait.dialogWait(*args, **kwargs)

    # Close the dialogWait
    def dialogWaitClose(self):
        """Close the dialogWait. See :func:`~dialogWait.dialogWait.close`."""
        if not self.waitdlg is None:
            self.waitdlg.close()
            self.waitdlg = None
        
        
    # Open a fab button
    def fab(self, *args, **kwargs):
        """Open a fab button. See :func:`~fab.fab` for the list of parameters."""
        kwargs['output'] = self.outfabs
        return fab.fab(*args, **kwargs)
        
            
    # Direct download of a .txt file containing a string
    def downloadText(self, textobj, fileName="download.txt"):
        """Direct download of a .txt file containing a string.
        
        Parameters
        ----------
        textobj : str
            Text to write in the file to be downloaded
        fileName : str, optional
            Name of the file that is to be downloaded (default is 'download.txt')

        Example
        -------
        If g_app is an instance of the app class, this code will download a text file containing the passed string::

            g_app.downloadText('aaa bbb ccc')
        """

        string_bytes  = textobj.encode("ascii","ignore")
        base64_bytes  = base64.b64encode(string_bytes)
        base64_string = base64_bytes.decode("ascii")

        self.outdownload.clear_output()
        with self.outdownload:
            display(HTML('<script>function downloadURI(uri, name) { var link = document.createElement("a"); link.download = name; link.href = uri; link.click();} downloadURI("data:application/octet-stream;charset=utf-8;base64,' + base64_string + '","' + fileName + '"); </script>'))
        self.outdownload.clear_output()


    # Direct download of an array of bytes
    def downloadBytes(self, bytesobj, fileName="download.bin"):
        """Direct download of an array of bytes.
        
        Parameters
        ----------
        bytesobj : bytes-like object
            Bytes to write in the file to be downloaded
        fileName : str, optional
            Name of the file that is to be downloaded (default is 'download.bin')

        Example
        -------
        If g_app is an instance of the app class, this code will download a binary file containing the passed bytes::

            g_app.downloadBytes(b'ajgh lkjhl ')
        """
        base64_bytes  = base64.b64encode(bytesobj)
        base64_string = base64_bytes.decode("ascii")

        self.outdownload.clear_output()
        with self.outdownload:
            display(HTML('<script>function downloadURI(uri, name){ var link = document.createElement("a"); link.download = name; link.href = uri; link.click();} downloadURI("data:application/octet-stream;charset=utf-8;base64,''' + base64_string + '","' + fileName + '");</script>'))
        self.outdownload.clear_output()
            
            
    # Open a web page in another tab
    def urlOpen(self, url, target='_blank'):
        """Open a web page in another tab.
        
        Parameters
        ----------
        url : str
            URL of the page to be opened
        target : str, optional
            Target of the open operation (default is '_blank' which means that the page will be opened in a new browser's tab)

        Example
        -------
        If g_app is an instance of the app class, this code will open a new tab in the browser::

            g_app.urlOpen('https://www.google.com')
        """
        js = '<script type=\"text/javascript\">window.open("%s", "%s");</script>' % (url,target)
        with self.outurlopen:
            display(ipyhtml(js))
                
                
    # Read the parameters passed on the URL 
    def urlParameter(self, parameterName, parameterDefaultValue=''):
        """Read the parameters passed on the URL.
        
        Parameters
        ----------
        parameterName : str
            name of the parameter to read from the URL that launched the Voilà dashboard
        parameterDefaultValue : str, optional
            Default value of the parameter, in case it is not present in the URL that launched the Voilà dashboard (default is '')

        Example
        -------
        If g_app is an instance of the app class, this code print the value of an URL parameter::

            print(g_app.urlParameter('activetab'))
        """
        return self.parameters.get(parameterName, [parameterDefaultValue])[0]
        
            
    # Update the visualized URL in the browser
    def urlUpdate(self, url):
        """Update the URL visualized in the top bar of the browser.
        
        Parameters
        ----------
        url : str
            Partial url to add to the current browser's page key/values

        Example
        -------
        If g_app is an instance of the app class, this code will add a key=value to the URL of the application::

            g_app.urlUpdate('?test=3')
        """
        
        js = "<script>window.history.replaceState({ additionalInformation: 'Updated the URL with JS' }, '', '%s');</script>" % url
        with self.outurlupdate:
                display(HTML(js))



"""Utility functions for downloading text and binary files"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

from ipywidgets import widgets, Layout
from IPython.display import display, HTML
import base64


# Output widget that needs to be visualized inside the notebook in order for the download to work!!!
output = widgets.Output(layout=Layout(width='0px', min_width='0px', height='0px', min_height='0px'))


# Direct download of a .txt file containing a string
def downloadText(textobj, fileName="download.txt"):
    """
    Download of a string as a text file.
    
    Parameters
    ----------
        textobj : str
            Text to be written in the text file to be downloaded
        fileName : str, optional
            Name of the file to download (default is "download.txt")
            
    Example
    -------
    In order to download a text file, the Output widget download.output must be displayed inside the notebook. This is required because the download operation is based on the execution of Javascript code, and this requires an Output widget displayed. After the download.output widget is visible, then the download.downloadText function can be called::
    
        from vois import download
        
        display(download.output)
        
        with download.output:
            download.downloadText('aaa bbb ccc')
    
    """

    string_bytes  = textobj.encode("ascii","ignore")
    base64_bytes  = base64.b64encode(string_bytes)
    base64_string = base64_bytes.decode("ascii")

    output.clear_output()
    with output:
        display(HTML('\
<script>function downloadURI(uri, name) \
{\
    var link = document.createElement("a");\
    link.download = name;\
    link.href = uri;\
    link.click();\
}\
downloadURI("data:application/octet-stream;charset=utf-8;base64,' + base64_string + '","' + fileName + '");\
</script>'))

    output.clear_output()
        

# Direct download of an array of bytes
def downloadBytes(bytesobj, fileName="download.bin"):
    """
    Download of an array of bytes as a binary file.
    
    Parameters
    ----------
        bytesobj : bytearray or bytes object
            Bytes to be written in the binary file to be downloaded
        fileName : str, optional
            Name of the file to download (default is "download.bin")
            
    Example
    -------
    In order to download a binary file, the Output widget download.output must be displayed inside the notebook. This is required because the download operation is based on the execution of Javascript code, and this requires an Output widget displayed. After the download.output widget is visible, then the download.downloadBytes function can be called::
    
        from vois import download
        
        display(download.output)
        
        with download.output:
            download.downloadBytes(bytearray(b'ajgh lkjhl '))
    
    """

    base64_bytes  = base64.b64encode(bytesobj)
    base64_string = base64_bytes.decode("ascii")

    output.clear_output()
    with output:
        display(HTML('''
<script>function downloadURI(uri, name)
{
    var link = document.createElement("a");
    link.download = name;
    link.href = uri;
    link.click();
}
downloadURI("data:application/octet-stream;charset=utf-8;base64,''' + base64_string + '","' + fileName + '");</script>'))
        
    output.clear_output()



"""SVG drawings for general use."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from ipywidgets import HTML, widgets, Layout
from ipyevents import Event
from IPython.display import display
import math
from datetime import datetime
from textwrap import wrap

try:
    from . import colors
    from .vuetify import fontsettings
    from .vuetify import settings
except:
    import colors
    from vuetify import fontsettings
    from vuetify import settings


###########################################################################################################################################################################
# Returns a string containing the legend for a vector layer classification in SVG format
###########################################################################################################################################################################
def categoriesLegend(title, descriptions, width=200, elemHeight=38, bordercolor='black', borderWidth=2, textcolor='black',
                     colorlist=['#fef4ef', '#fdd3c0', '#fca080', '#fa6a49', '#e23026', '#b11117', '#66000c'],
                     dark=settings.dark_mode):
    """
    Creation of a legend for categories given title and descriptions of the classes.
    
    Parameters
    ----------
    title : str
        Title of the legend
    descriptions : list of strings
        Description of each of the classe of the legend. If a description contains a '\t' character, the text after it is displayed as multi-line text in smaller font
    width : int, optional
        Width in pixel of the legend (default is 200)
    elemHeight : int, optional
        Height in pixels of each of the  elements of the legend (default is 38)
    bordercolor : str, optional
        Color of the border of the legend elements (default is 'black')
    borderWidth : int, optional
        Width in pixels of the border of the legend elements (default is 2)
    textcolor : str, optional
        Color of text for the legend elements (default is 'black')
    colorlist : list of colors, optional
        List of colors to assign to the country polygons (see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )
    dark : bool, optional
        If True, the bordercolor and textcolor are set to white (default is False)
    
    Return
    ------
        a string containing SVG text to display the legend
    
    Example
    -------
    Example of the creation of an SVG drawing for a categories legend::
    
        from vois import svgUtils
        import plotly.express as px

        svg = svgUtils.categoriesLegend("Legend title",
                                        ['1:\tVery long class description that can span multiple lines and that contains no info at all',
                                         'Class 2', 'Class 3', 'Class 4'],
                                        colorlist=px.colors.sequential.Blues,
                                        width=250)
        display(HTML(svg))
        
    .. figure:: figures/categoriesLegend.png
       :scale: 100 %
       :alt: categoriesLegend example

       Example of a categories legend

    """
    
    def multilineText(x,y,width,height, text):
        return '''
<switch>
   <foreignObject x="%d" y="%d" width="%d" height="%d">
     <div style="display: table-cell; vertical-align: middle; width: %dpx; height: %dpx;">
       <p xmlns="http://www.w3.org/1999/xhtml" style="line-height: 14px; font-size: 8pt;">%s</p>
     </div>
   </foreignObject>
   <text x="%d" y="%d">Your SVG viewer cannot display html.</text>
  <title>%s</title>
</switch>''' % (x,y,width,height,width,height, text, x,y, text)
    
    if dark:
        if bordercolor=='black': bordercolor='white'
        if textcolor  =='black': textcolor  ='white'
            
    x = 0
    y = 12
    space = 4
    n = len(descriptions)
    svg = '<svg xmlns="http://www.w3.org/2000/svg" width="%d" height="%d" version="1.1">\n' % (width, 28+n*(elemHeight+space))
    
    svg += '<text text-anchor="start" x="%d" y="%d" font-size="15" fill="%s" font-weight="500">%s</text>' % (x, y+5, textcolor, title)
    y += elemHeight/2
    
    s = colors.colorInterpolator(colorlist, 1, max(2,n))
    for i in range(n):
        col = s.GetColor(i+1)
        text = descriptions[i]
        if '\t' in text:
            value = text.split('\t')[0]
            text  = text.split('\t')[1]
            svg += '<rect style="fill:%s;" x="%d" y="%d" width="%d" height="%d" stroke="%s" stroke-width="%d"><title>%s</title></rect>\n' % (col, x+1, y, elemHeight-1, elemHeight-1, bordercolor, borderWidth, value)
            svg += '<text text-anchor="start" x="%d" y="%d" font-size="15" fill="%s" font-weight="400">%s<title>%s</title></text>' % (x+elemHeight+space, y+elemHeight/2+4, textcolor, value, text)
            if len(text) > 0: svg += multilineText(x+2*elemHeight,y-1,width-2*elemHeight-space,elemHeight+10, text)
        else:
            value = text
            svg += '<rect style="fill:%s;" x="%d" y="%d" width="%d" height="%d" stroke="%s" stroke-width="%d"><title>%s</title></rect>\n' % (col, x+1, y, elemHeight-1, elemHeight-1, bordercolor, borderWidth, text)
            svg += '<text text-anchor="start" x="%d" y="%d" font-size="15" fill="%s" font-weight="400">%s<title>%s</title></text>' % (x+elemHeight+space, y+elemHeight/2+4, textcolor, value, text)
                
        y += elemHeight + space

    svg += '</svg>'
    return svg






###########################################################################################################################################################################
# Returns svg string containing a vertical graduated/colors legend
###########################################################################################################################################################################
def graduatedLegend(df,                          # Pandas dataframe indexed on code_column and containing 'value' and 'label' columns
                    code_column=None,            # Name of the column containing the code of the country (None = the country code is the index of the dataframe)
                    value_column='value',        # Name of the column containing the value
                    label_column='label',        # Name of the column containing the label
                    dictnames=None,              # Dict to convert codes to names
                    codes_selected=[],           # codes of the countries selected
                    colorlist=['#0d0887', '#46039f', '#7201a8', '#9c179e', '#bd3786', '#d8576b', '#ed7953', '#fb9f3a', '#fdca26', '#f0f921'],   # default color scale
                    stdevnumber=2.0,             # Number of stddev to calculate (minvalue,maxvalue) range
                    fill='#f1f1f1',              # fill color for countries
                    stroke_selected='#00ffff',   # stroke color for border of selected country
                    decimals=2,                  # Number of decimals for the legend number display
                    minallowed_value=None,       # Minimum value allowed
                    maxallowed_value=None,       # Maximum value allowed
                    hoveronempty=False,          # If True highlights polygon on hover even if no value present in input df for the polygon
                    legendtitle='',              # Title to add to the legend (top)
                    legendunits='',              # Units of measure to add to the legend (bottom)
                    fontsize=20,
                    width=200,
                    height=600,
                    bordercolor='black',
                    textcolor='black',
                    dark=False):
    """
    Creation of graduated legend in SVG format. Given a Pandas DataFrame in the same format of the one in input to :py:func:`interMap.geojsonMap` function, this functions generates an SVG drawing displaying a graduated colors legend. 

    Parameters
    ----------
    df : Pandas DataFrame
        Pandas DataFrame to use for assigning values to features. It has to contain at least a column with numeric values.
    code_column : str, optional
        Name of the column of the df Pandas DataFrame containing the unique code of the features. This column is used to perform the join with the internal attribute of the geojson vector dataset that contains the unique code. If the code_column is None, the code is taken from the index of the DataFrame, (default is None)
    value_column : str, optional
        Name of the column of the Pandas DataFrame containing the values to be assigned to the features using the join on geojson unique codes (default is 'value')
    label_column : str, optional
        Name of the column of the Pandas DataFrame containing the label to be assigned to the features using the join on geojson unique codes (default is 'label')
    dictnames : dict, optional
        Dictionary to convert codes to names when displaying the selection (default is None)
    codes_selected : list of strings, optional
        List of codes of features to display as selected (default is [])
    colorlist : list of colors, optional
        List of colors to assign to the country polygons (default is the Plotly px.colors.sequential.Plasma, see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )
    stdevnumber : float, optional
        The correspondance between the values assigned to features and the colors list is done by calculating a range of values [min,max] to linearly map the values to the colors. This range is defined by calculating the mean and standard deviation of the country values and applying this formula [mean - stdevnumber*stddev, mean + stdevnumber*stddev]. Default is 2.0
    fill : str, optional
        Fill color to use for the features that are not joined (default is '#f1f1f1')
    stroke_selected : str, optional
        Color to use for the selected features (default is '#00ffff')
    decimals : int, optional
        Number of decimals for the legend numbers display (default is 2)
    minallowed_value : float, optional
        Minimum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    maxallowed_value : float, optional
        Maximum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    hoveronempty : bool, optional
        If True highlights polygon on hover even if no value present in input df for the feature (default is False)
    legendtitle : str, optional
        Title to add on top of the legend (default is '')
    legendunits : str, optional
        Units of measure to add to the bottom of the legend (default is '')
    fontsize : int, optional
        Size in pixels of the font used for texts (default is 20)
    width : int, optional
        Width of the SVG drawing in pixels (default is 200)
    height : int, optional
        Height of the SVG drawing in pixels (default is 600)
    bordercolor : str, optional
        Color for lines and rects of the legend (default is 'black')
    textcolor : str, optional
        Color for texts of the legend (default is 'black')
    dark : bool, optional
        If True, the bordercolor and textcolor are set to white (default is False)
        
    Returns
    -------
        a string containing SVG text to display the graduated legend

    Example
    -------
    Creation of a SVG drawing to display a graduated legend. Input is prepared in the same way of the example provided for the :py:func:`interMap.geojsonMap` function::
        
        import numpy as np
        import pandas as pd
        import plotly.express as px
        from vois import svgMap, svgUtils

        countries = svgMap.country_codes

        # Generate random values and create a dictionary: key=countrycode, value=random in [0.0,100.0]
        d = dict(zip(countries, list(np.random.uniform(size=len(countries),low=0.0,high=100.0))))

        # Create a pandas dataframe from the dictionary
        df = pd.DataFrame(d.items(), columns=['iso2code', 'value'])

        svg = svgUtils.graduatedLegend(df, code_column='iso2code',
                                       codes_selected=['IT', 'FR', 'CH'],
                                       stroke_selected='red',
                                       colorlist=px.colors.sequential.Viridis[::-1],
                                       stdevnumber=2.0,
                                       legendtitle='2020 Total energy consumption',
                                       legendunits='KTOE per 100K inhabit.',
                                       fontsize=18,
                                       width=340, height=600)
        display(HTML(svg))

    .. figure:: figures/graduatedLegend.png
       :scale: 100 %
       :alt: graduatedLegend example

       Example of a graduatedLegend in SVG
        
    """

    if df.shape[0] <= 0:
        minvalue = 1.0
        maxvalue = 2.0
    else:
        mean = df[value_column].mean()
        if df.shape[0] <= 1:
            minvalue = mean
            maxvalue = mean
        else:
            stddev = df[value_column].std()
            valuemin = df[value_column].min()
            valuemax = df[value_column].max()

            minvalue = mean - stdevnumber*stddev
            maxvalue = mean + stdevnumber*stddev

            if minvalue < valuemin: minvalue = valuemin
            if maxvalue > valuemax: maxvalue = valuemax

        if not minallowed_value is None:
            if minvalue < minallowed_value: minvalue = minallowed_value
        if not maxallowed_value is None:
            if maxvalue > maxallowed_value: maxvalue = maxallowed_value
        
    #print(mean,stddev, minvalue, maxvalue)

    if minvalue >= maxvalue: maxvalue = minvalue + 1
    ci = colors.colorInterpolator(colorlist,minvalue,maxvalue)

    if dark:
        if bordercolor=='black': bordercolor='white'
        if textcolor  =='black': textcolor  ='white'
    
    # Positioning of the legend
    w = width // 5
    x1 = (width-w)/2
    x2 = x1 + w
    wlineette = width // 20
    
    #y1 = height // 18
    y1 = 2*fontsize
    h = height - int(1.5*y1)
    if len(legendunits) > 0:
        h -= y1//2
    y2 = y1 + h
    
    barthickness = width/120
    fontsize1 = fontsize - 1
    fontsize2 = fontsize - 2
    
    # Calculate the legend in SVG format
    svg = '''
<svg xmlns:svg="http://www.w3.org/2000/svg" xmlns="http://www.w3.org/2000/svg" width="%d" height="%d" version="1.1">
  <style type="text/css">
     @import url('%s');
''' % (width,height,fontsettings.font_url)

    # Colors indexed by iso2_code of countries
    polycolors = {}    # Fill color of the polygon
    polyclass  = {}    # class to assign to the polygon ("country" or "")
    polybary   = {}    # Y coordinate of the bar on the legend highlighted when hover on a country
    polyover   = {}    # True if a value is outside of the legend
    polyname   = {}    # Name assigned to the polygon
    
    country_codes = df[code_column].unique()
    
    for c in country_codes:
        polycolors[c] = fill
        polybary[c]   = -1000
        polyover[c]   = False
        
        if hoveronempty:
            polyclass[c] = 'country'
        else:
            polyclass[c] = ''
        
    # Set colors for all the countries
    for index, row in df.iterrows():
        if code_column is None: code = index
        else:                   code = row[code_column]

        polyname[code] = code
        if not dictnames is None and code in dictnames:
            polyname[code] = dictnames[code]
            
        value = row[value_column]
        polycolors[code] = ci.GetColor(value)
        polyclass[code]  = 'country'

        if label_column in df:
            v = row[label_column]
        else:
            v = str(value)
                
        y = y2 - (y2-y1) * (value - minvalue) / (maxvalue - minvalue)
        if y < y1: y = y1
        if y > y2: y = y2
        polybary[code] = y
        
        if value < minvalue or value > maxvalue:
            polyover[code] = True
            
    #print(minvalue,maxvalue, y1, y2)
        
    # Add color for every polygon
    for c in country_codes:
        svg += 'svg #%s { fill: %s; }\n' % (c, polycolors[c])
        
    svg += '</style>'

    
    # Legend on the right
    if len(legendtitle) > 0:
        svg += '<text x="%d" y="%d" text-anchor="middle" font-size="%f" font-family="%s" font-weight="bold" fill="%s">%s</text>' % (x1+w/2.0, y1-fontsize, fontsize, fontsettings.font_name, textcolor, legendtitle)
        
    if len(legendunits) > 0:
        svg += '<text x="%d" y="%d" text-anchor="middle" font-size="%f" font-family="%s" font-weight="bold" fill="%s">%s</text>' % (x1+w/2.0, int(y2+fontsize*1.5), fontsize2, fontsettings.font_name, textcolor, legendunits)
        
    svg += '<rect x="%d" y="%d" width="%d" height="%d" style="fill:none; stroke-width:%f; stroke:%s;" />' % (x1, y1, w, h+1, barthickness*2, bordercolor)
    
    y = y2
    for i in range(h):
        value = maxvalue - (y - y1) * (maxvalue - minvalue) / (y2 - y1)
        svg += '<line x1="%d" y1="%d" x2="%d" y2="%d" style="stroke:%s;stroke-width:%f" />' % ( x1,y,x2,y, ci.GetColor(value), barthickness )
        y -= 1
        
    svg += '<line x1="%d" y1="%d" x2="%d" y2="%d" style="stroke:%s; stroke-width:%f" />' % ( x2,y2+3,x2+wlineette,y2+3, bordercolor, barthickness/2.0 )
    svg += '<line x1="%d" y1="%d" x2="%d" y2="%d" style="stroke:%s; stroke-width:%f" />' % ( x2,y1-2,x2+wlineette,y1-2, bordercolor, barthickness/2.0 )
    
    valmin = '{:.{prec}f}'.format(minvalue, prec=decimals)
    svg += '<text x="%d" y="%d" font-size="%f" font-family="%s" fill="%s">%s</text>' % (x2+wlineette+5, y2+fontsize2/3, fontsize1, fontsettings.font_name, textcolor, valmin)
    
    valmax = '{:.{prec}f}'.format(maxvalue, prec=decimals)
    svg += '<text x="%d" y="%d" font-size="%f" font-family="%s" fill="%s">%s</text>' % (x2+wlineette+5, int(y1+fontsize2*0.4), fontsize1, fontsettings.font_name, textcolor, valmax)
    
    valmed = '{:.{prec}f}'.format((minvalue+maxvalue)/2.0, prec=decimals)
    if valmed != valmin and valmed != valmax:
        y = (y1+y2)/2.0
        svg += '<line x1="%d" y1="%f" x2="%d" y2="%f" style="stroke:%s; stroke-width:%f" />' % ( x2,y,x2+wlineette,y, bordercolor, barthickness/2.0 )
        svg += '<text x="%d" y="%d" font-size="%f" font-family="%s" fill="%s">%s</text>' % (x2+wlineette+5, y+fontsize2/3, fontsize1, fontsettings.font_name, textcolor, valmed)
        
    val = '{:.{prec}f}'.format(minvalue + 3.0*(maxvalue-minvalue)/4.0, prec=decimals)
    if val != valmed and val != valmax:
        y = y1+(y2-y1)/4.0
        svg += '<line x1="%d" y1="%f" x2="%d" y2="%f" style="stroke:%s; stroke-width:%f" />' % ( x2,y,x2+wlineette,y, bordercolor, barthickness/2.0 )
        svg += '<text x="%d" y="%d" font-size="%f" font-family="%s" fill="%s">%s</text>' % (x2+wlineette+5, y+fontsize2/3, fontsize1, fontsettings.font_name, textcolor, val)
        
    val = '{:.{prec}f}'.format(minvalue + (maxvalue-minvalue)/4.0, prec=decimals)
    if val != valmin and val != valmed:
        y = y1+3.0*(y2-y1)/4.0
        svg += '<line x1="%d" y1="%f" x2="%d" y2="%f" style="stroke:%s; stroke-width:%f" />' % ( x2,y,x2+wlineette,y, bordercolor, barthickness/2.0 )
        svg += '<text x="%d" y="%d" font-size="%f" font-family="%s" fill="%s">%s</text>' % (x2+wlineette+5, y+fontsize2/3, fontsize1, fontsettings.font_name, textcolor, val)
    
    
    # Add horizontal lines in the legend for the selected countries
    for code in codes_selected:
        if code in country_codes:
            if polybary[code] >= y1 and polybary[code] <= y2:
                svg += '<text x="%f" y="%f" text-anchor="end" font-size="%f" font-family="%s" font-weight="bold" fill="%s">%s<title>%s</title></text>' % (x1-wlineette/2, polybary[code]+fontsize2/3, fontsize2, fontsettings.font_name, textcolor, polyname[code], polyname[code])

                dash = ''
                if polyover[code]:
                    dash = 'stroke-dasharray="%f,%f"' % (wlineette*1.1, wlineette*0.5)
                svg += '<line x1="%d" y1="%d" x2="%d" y2="%d" style="stroke:%s; stroke-width:%f" %s />' % (x1, polybary[code], x2, polybary[code], stroke_selected, barthickness, dash)
    
    svg += '</svg>'
    return svg




###########################################################################################################################################################################
# Returns svg string containing a vertical graduated/colors legend
###########################################################################################################################################################################
def graduatedLegendVWVH(df,                          # Pandas dataframe indexed on code_column and containing 'value' and 'label' columns
                    code_column=None,            # Name of the column containing the code of the country (None = the country code is the index of the dataframe)
                    value_column='value',        # Name of the column containing the value
                    label_column='label',        # Name of the column containing the label
                    codes_selected=[],           # codes of the countries selected
                    colorlist=['#0d0887', '#46039f', '#7201a8', '#9c179e', '#bd3786', '#d8576b', '#ed7953', '#fb9f3a', '#fdca26', '#f0f921'],   # default color scale
                    stdevnumber=2.0,             # Number of stddev to calculate (minvalue,maxvalue) range
                    fill='#f1f1f1',              # fill color for countries
                    stroke_selected='#00ffff',   # stroke color for border of selected country
                    decimals=2,                  # Number of decimals for the legend number display
                    minallowed_value=None,       # Minimum value allowed
                    maxallowed_value=None,       # Maximum value allowed
                    hoveronempty=False,          # If True highlights polygon on hover even if no value present in input df for the polygon
                    legendtitle='',              # Title to add to the legend (top)
                    legendunits='',              # Units of measure to add to the legend (bottom)
                    fontsize=20,
                    width=20.0,
                    height=40.0,
                    bordercolor='black',
                    textcolor='black',
                    dark=False):
    """
    Creation of graduated legend in SVG format. Given a Pandas DataFrame in the same format of the one in input to :py:func:`interMap.geojsonMap` function, this functions generates an SVG drawing displaying a graduated colors legend. 

    Parameters
    ----------
    df : Pandas DataFrame
        Pandas DataFrame to use for assigning values to features. It has to contain at least a column with numeric values.
    code_column : str, optional
        Name of the column of the df Pandas DataFrame containing the unique code of the features. This column is used to perform the join with the internal attribute of the geojson vector dataset that contains the unique code. If the code_column is None, the code is taken from the index of the DataFrame, (default is None)
    value_column : str, optional
        Name of the column of the Pandas DataFrame containing the values to be assigned to the features using the join on geojson unique codes (default is 'value')
    label_column : str, optional
        Name of the column of the Pandas DataFrame containing the label to be assigned to the features using the join on geojson unique codes (default is 'label')
    codes_selected : list of strings, optional
        List of codes of features to display as selected (default is [])
    colorlist : list of colors, optional
        List of colors to assign to the country polygons (default is the Plotly px.colors.sequential.Plasma, see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )
    stdevnumber : float, optional
        The correspondance between the values assigned to features and the colors list is done by calculating a range of values [min,max] to linearly map the values to the colors. This range is defined by calculating the mean and standard deviation of the country values and applying this formula [mean - stdevnumber*stddev, mean + stdevnumber*stddev]. Default is 2.0
    fill : str, optional
        Fill color to use for the features that are not joined (default is '#f1f1f1')
    stroke_selected : str, optional
        Color to use for the border of the selected features (default is '#00ffff')
    decimals : int, optional
        Number of decimals for the legend numbers display (default is 2)
    minallowed_value : float, optional
        Minimum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    maxallowed_value : float, optional
        Maximum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    hoveronempty : bool, optional
        If True highlights polygon on hover even if no value present in input df for the feature (default is False)
    legendtitle : str, optional
        Title to add on top of the legend (default is '')
    legendunits : str, optional
        Units of measure to add to the bottom of the legend (default is '')
    fontsize : int, optional
        Size in pixels of the font used for texts (default is 20)
    width : float, optional
        Width of the SVG drawing in vw units (default is 20.0)
    height : float, optional
        Height of the SVG drawing in vh units (default is 40.0)
    bordercolor : str, optional
        Color for lines and rects of the legend (default is 'black')
    textcolor : str, optional
        Color for texts of the legend (default is 'black')
    dark : bool, optional
        If True, the bordercolor and textcolor are set to white (default is False)
        
    Returns
    -------
        a string containing SVG text to display the graduated legend

    Example
    -------
    Creation of a SVG drawing to display a graduated legend. Input is prepared in the same way of the example provided for the :py:func:`interMap.geojsonMap` function::
        
        import numpy as np
        import pandas as pd
        import plotly.express as px
        from vois import svgMap, svgUtils

        countries = svgMap.country_codes

        # Generate random values and create a dictionary: key=countrycode, value=random in [0.0,100.0]
        d = dict(zip(countries, list(np.random.uniform(size=len(countries),low=0.0,high=100.0))))

        # Create a pandas dataframe from the dictionary
        df = pd.DataFrame(d.items(), columns=['iso2code', 'value'])

        svg = svgUtils.graduatedLegend(df, code_column='iso2code',
                                       codes_selected=['IT', 'FR', 'CH'],
                                       stroke_selected='red',
                                       colorlist=px.colors.sequential.Viridis[::-1],
                                       stdevnumber=2.0,
                                       legendtitle='2020 Total energy consumption',
                                       legendunits='KTOE per 100K inhabit.',
                                       fontsize=18,
                                       width=340, height=600)
        display(HTML(svg))

    .. figure:: figures/graduatedLegend.png
       :scale: 100 %
       :alt: graduatedLegend example

       Example of a graduatedLegend in SVG
        
    """

    if df.shape[0] <= 0:
        minvalue = 1.0
        maxvalue = 2.0
    else:
        mean = df[value_column].mean()
        if df.shape[0] <= 1:
            minvalue = mean
            maxvalue = mean
        else:
            stddev = df[value_column].std()
            valuemin = df[value_column].min()
            valuemax = df[value_column].max()

            minvalue = mean - stdevnumber*stddev
            maxvalue = mean + stdevnumber*stddev

            if minvalue < valuemin: minvalue = valuemin
            if maxvalue > valuemax: maxvalue = valuemax

        if not minallowed_value is None:
            if minvalue < minallowed_value: minvalue = minallowed_value
        if not maxallowed_value is None:
            if maxvalue > maxallowed_value: maxvalue = maxallowed_value
        
    #print(mean,stddev, minvalue, maxvalue)

    if minvalue >= maxvalue: maxvalue = minvalue + 1
    ci = colors.colorInterpolator(colorlist,minvalue,maxvalue)

    if dark:
        if bordercolor=='black': bordercolor='white'
        if textcolor  =='black': textcolor  ='white'

    # Sizable dimensioning
    svgwidth = 100.0
    aspectratio = 0.5*height / width   # In landscape mode, usually the height is half the width dimension!!!
    svgheight = svgwidth * aspectratio
            
    # Positioning of the legend
    w = svgwidth / 5.0
    x1 = (svgwidth-w)/2
    x2 = x1 + w
    wlineette = svgwidth / 25.0
    
    #y1 = height // 18
    y1 = 2*fontsize
    h = svgheight - 1.5*y1
    if len(legendunits) > 0:
        h -= y1/2.0
    y2 = y1 + h
    
    barthickness = width/50.0
    fontsize1 = fontsize*0.92
    fontsize2 = fontsize*0.85
    
    # Calculate the legend in SVG format
    preserve = 'xMidYMid meet'
    svg = '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" xml:space="preserve" viewBox="0 0 %f %f" preserveAspectRatio="%s" width="%fvw" height="%fvh">' % (svgwidth,svgheight, preserve, width,height)
    
    svg += '''
  <style type="text/css">
     @import url('%s');
''' % fontsettings.font_url

    # Colors indexed by iso2_code of countries
    polycolors = {}    # Fill color of the polygon
    polyclass  = {}    # class to assign to the polygon ("country" or "")
    polybary   = {}    # Y coordinate of the bar on the legend highlighted when hover on a country
    
    country_codes = df[code_column].unique()
    
    for c in country_codes:
        polycolors[c] = fill
        polybary[c]   = -1000
        if hoveronempty:
            polyclass[c] = 'country'
        else:
            polyclass[c] = ''
        
    # Set colors for all the countries
    for index, row in df.iterrows():
        if code_column is None: code = index
        else:                   code = row[code_column]
        value = row[value_column]
        polycolors[code] = ci.GetColor(value)
        polyclass[code]  = 'country'

        if label_column in df:
            v = row[label_column]
        else:
            v = str(value)
                
        y = y2 - (y2-y1) * (value - minvalue) / (maxvalue - minvalue)
        if y < y1: y = y1
        if y > y2: y = y2
        polybary[code] = y
        
    #print(minvalue,maxvalue, y1, y2)
        
    # Add color for every polygon
    for c in country_codes:
        svg += 'svg #%s { fill: %s; }\n' % (c, polycolors[c])
        
    svg += '</style>'

    
    # Legend on the right
    if len(legendtitle) > 0:
        svg += '<text x="%d" y="%d" text-anchor="middle" font-size="%f" font-family="%s" font-weight="bold" fill="%s">%s</text>' % (x1+w/2.0, y1-fontsize, fontsize, fontsettings.font_name, textcolor, legendtitle)
        
    if len(legendunits) > 0:
        svg += '<text x="%f" y="%f" text-anchor="middle" font-size="%f" font-family="%s" font-weight="bold" fill="%s">%s</text>' % (x1+w/2.0, y2+fontsize*1.5, fontsize2, fontsettings.font_name, textcolor, legendunits)
        
    svg += '<rect x="%f" y="%f" width="%f" height="%f" style="fill:none; stroke-width:%f; stroke:%s;" />' % (x1, y1, w, h, barthickness*4.0, bordercolor)
    
    y = y1
    while y <= y2:
        value = maxvalue - (y - y1) * (maxvalue - minvalue) / (y2 - y1)
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" style="stroke:%s;stroke-width:%f" />' % ( x1,y,x2,y, ci.GetColor(value), barthickness )
        y += 0.06666
        
    svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" style="stroke:%s; stroke-width:%f" />' % ( x2,y2,x2+wlineette,y2, bordercolor, barthickness/2.0 )
    svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" style="stroke:%s; stroke-width:%f" />' % ( x2,y1,x2+wlineette,y1, bordercolor, barthickness/2.0 )

    xtext = x2 + wlineette*1.25
    
    valmin = '{:.{prec}f}'.format(minvalue, prec=decimals)
    svg += '<text x="%f" y="%f" font-size="%f" font-family="%s" fill="%s">%s</text>' % (xtext, y2+fontsize2/3, fontsize1, fontsettings.font_name, textcolor, valmin)
    
    valmax = '{:.{prec}f}'.format(maxvalue, prec=decimals)
    svg += '<text x="%f" y="%f" font-size="%f" font-family="%s" fill="%s">%s</text>' % (xtext, int(y1+fontsize2*0.4), fontsize1, fontsettings.font_name, textcolor, valmax)
    
    valmed = '{:.{prec}f}'.format((minvalue+maxvalue)/2.0, prec=decimals)
    if valmed != valmin and valmed != valmax:
        y = (y1+y2)/2.0
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" style="stroke:%s; stroke-width:%f" />' % ( x2,y,x2+wlineette,y, bordercolor, barthickness/2.0 )
        svg += '<text x="%f" y="%f" font-size="%f" font-family="%s" fill="%s">%s</text>' % (xtext, y+fontsize2/3.0, fontsize1, fontsettings.font_name, textcolor, valmed)
        
    val = '{:.{prec}f}'.format(minvalue + 3.0*(maxvalue-minvalue)/4.0, prec=decimals)
    if val != valmed and val != valmax:
        y = y1+(y2-y1)/4.0
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" style="stroke:%s; stroke-width:%f" />' % ( x2,y,x2+wlineette,y, bordercolor, barthickness/2.0 )
        svg += '<text x="%f" y="%f" font-size="%f" font-family="%s" fill="%s">%s</text>' % (xtext, y+fontsize2/3.0, fontsize1, fontsettings.font_name, textcolor, val)
        
    val = '{:.{prec}f}'.format(minvalue + (maxvalue-minvalue)/4.0, prec=decimals)
    if val != valmin and val != valmed:
        y = y1+3.0*(y2-y1)/4.0
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" style="stroke:%s; stroke-width:%f" />' % ( x2,y,x2+wlineette,y, bordercolor, barthickness/2.0 )
        svg += '<text x="%f" y="%f" font-size="%f" font-family="%s" fill="%s">%s</text>' % (xtext, y+fontsize2/3.0, fontsize1, fontsettings.font_name, textcolor, val)
    
    
    # Add horizontal lines in the legend for the selected countries
    for code in codes_selected:
        if code in country_codes:
            if polybary[code] >= y1 and polybary[code] <= y2:
                svg += '<text x="%f" y="%f" text-anchor="end" font-size="%f" font-family="%s" font-weight="bold" fill="%s">%s</text>' % (x1-wlineette/2, polybary[code]+fontsize2/3, fontsize2, fontsettings.font_name, textcolor, code)
                svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" style="stroke:%s; stroke-width:%f" />' % (x1, polybary[code], x2, polybary[code], stroke_selected, barthickness*3.0)
    
    svg += '</svg>'
    return svg


###########################################################################################################################################################################
# https://www.smashingmagazine.com/2019/01/html5-svg-fill-animation-css3-vanilla-javascript/
# Small circle with animation: returns a widgets.Output object with the SVG displayed on it
###########################################################################################################################################################################
def SmallCircle(text1, text2, percentage, forecolor="#308040", backcolor=None, textcolor='white', dimension=300.0, fontsize=16.0, textsize=0.0):
    """
    Display of a circle graphics displaying a text and a percentage value. It shows an animation to reach the requested percentage of the full circle.
    
    Parameters
    ----------
    text1 : str
        First string of text to display inside the circle (usually a brief description of the variable displayed)
    text2 : str
        Second string of text to display inside the circle (usually the numercal value)
    percentage : float
        Value in percentage to be displayed
    forecolor : str, optional
        Color to use for the circle border (default is '#308040')
    backcolor : str, optional
        Color to use for the interior of the circle (default is None, so a lighter color is generated from the forecolor)
    textcolor : str, optional
        Color to use for text (default is 'white')
    dimension : int or float or str, optional
        Side of the drawing. If an integer or a float is passed, the size is intended in pixels units, otherwise a string containing the units must be passed (example: '4vh'). The default is 300.0 for 300 pixels
    textsize : float, optional
        Text dimension in pixels (default is 0.0 which means that it is automatically calculated from the dimension of the drawing)
        
    Returns
    -------
        an ipywidgets.Output instance with the circle already displayed inside
        
    Example
    -------
    Example of a circle to represent a percentage with an animation::
    
        from vois import svgUtils
        from random import randrange

        percentage = randrange(1000)/10.0
        svgUtils.SmallCircle('Green<br>deal',
                             '%.1f%%' % percentage,
                             percentage,
                             forecolor="#308040",
                             dimension=200)    
    
    .. figure:: figures/smallCircle.png
       :scale: 100 %
       :alt: smallCircle example

       Example of an animated SVG to graphically represent a percentage value
    
    """

    def circumference(r):
        return 2.0 * r * math.pi
    
    def complementaryColor(color):
        if color[0] == '#':
            color = color[1:]
        rgb = (color[0:2], color[2:4], color[4:6])
        comp = ['%02X' % (255 - int(a, 16)) for a in rgb]
        return '#' + ''.join(comp)

    def lighterColor(color):
        if color[0] == '#':
            color = color[1:]
        rgb = (color[0:2], color[2:4], color[4:6])
        comp = ['%02X' % min(255,int(int(a, 16)*1.5)) for a in rgb]
        return '#' + ''.join(comp)

    if backcolor is None:
        backcolor = lighterColor(forecolor)
        
    tooltip = text1.replace('<br>',' ') + ': ' + text2
    if percentage < 0.0:   percentage = 0.0
    if percentage > 100.0: percentage = 100.0
    
    r = 44
    circle = circumference(r)
    #value = 276.46015351590177  # empty
    #value = 0.0                 # full
    value = 0.01 * (100.0 - percentage)*circle
    seconds = 1.0 + (circle - value) / circle
    
    svg = '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" x="0px" y="0px" viewBox="0 0 100 100" xml:space="preserve">'

    # CSS styling
    svg += '''
        <style type="text/css">
           @import url('%s');
        </style>
    ''' % (fontsettings.font_url)

        
    svg += '<circle fill="%s" cx="50" cy="50" r="44"><title>%s</title></circle>' % (backcolor,tooltip)
    svg += '<circle fill="none" stroke="#f1f1f1" stroke-width="10" stroke-mitterlimit="0" cx="50" cy="50" r="%d"><title>%s</title></circle>' % (r,tooltip)
    svg += '''
        <circle fill="none" stroke="%s" stroke-width="10.5" stroke-mitterlimit="0" cx="50" cy="50" r="%d" stroke-dasharray="%d" stroke-dashoffset="%d" stroke-linecap="butt" transform="rotate(-90 ) translate(-100 0)">
            <animate attributeName="stroke-dashoffset" values="%d;%d;%d" dur="%fs"></animate>
            <title>%s</title>
        </circle>
    ''' % (forecolor, r, circle, value, circle, value, value, seconds, tooltip)
    

    vtext = text1.split('<br>')
    maxlen = 0
    for t in vtext:
        if len(t) > maxlen: maxlen = len(t)

    if textsize <= 0:  size = 1.0 * 10.0 / float(maxlen)
    else:              size = textsize
        
    if size > 0.9: size = 0.9
    if size < 0.4: size = 0.4
    
    h = int(14 * size)
    y = 47 - (len(vtext)-1)*h
    for t in vtext:
        svg += '<text font-size="%.2fem" text-anchor="middle" x="50" y="%d" fill="%s" font-weight="bold" style="font-family: %s;">%s<title>%s</title></text>' % (size,y, textcolor, fontsettings.font_name,t,tooltip)
        y += h
        
    svg += '<text font-size="1em" text-anchor="middle" x="50" y="62" fill="%s" font-weight="bold" style="font-family: %s;">%s<title>%s</title></text>' % (textcolor, fontsettings.font_name, text2, tooltip)
    
    svg += '</svg>'
    
    # Create an output widget and display SVG in it
    if isinstance(dimension, int):
        w = '%dpx' % dimension
    elif isinstance(dimension, float):
        w = '%fpx' % dimension
    else:
        w = str(dimension)
    out = widgets.Output(layout=Layout(width=w, height='calc(%s + 14px)' % w))

    with out:
        display(HTML(svg))
        
    return out



###########################################################################################################################################################################
# https://mirellavanteulingen.nl/blog/svg-charts-animations.html
# Pie chart with animation: returns a widgets.Output object with the SVG displayed on it + the svg text 
###########################################################################################################################################################################

# Enumeration
class FillPercentage():
    fill100 = 0
    fill75  = 1
    fill60  = 2
    fill50  = 3
    fill40  = 4

def AnimatedPieChart(values=[10.0, 25.0, 34.0, 24.0, 23.0], colors=['#2d82c2', '#95cb92', '#e7ee99', '#ffde88', '#ff945a', '#e34e4f'], labels=None, duration=0.75, # Seconds
                     fillpercentage=FillPercentage.fill60, backcolor="#f1f1f1", dimension=400,
                     textcolor=settings.select_textcolor, fontsize=15, textweight=400, decimals=1,
                     centertext='', centercolor=settings.select_textcolor, centerfontsize=18, centertextweight=500,
                     onclick=None, additional_argument=None, is_selected=False, displayvalues=True):
    """
    Creation of an animated pie chart in SVG format. Given an array of float values, and optional labels, the function draws a pie chart that fills its slices with a short animation. An ipywidgets.Output instance is returned, which has the SVG chart displayed in it. By passing a value to the onclick parameter, it is possible to manage the click event on the slices of the pie, providing interactivity to the drawing. The capture of the click event is done using the `ipyevents library <https://github.com/mwcraig/ipyevents>`_ .
    
    Parameters
    ----------
    values : list of float values, optional
        List of float values that represent the relative dimension of each of the slices (default is [10.0, 25.0, 34.0, 24.0, 23.0])
    colors : list of strings representing colors, optional
        Colors to use for each of the slices of the pie (default is ['#2d82c2', '#95cb92', '#e7ee99', '#ffde88', '#ff945a', '#e34e4f'])
    labels, list of strings, optional
        Labels for each of the slices of the pie (default is None)
    duration : float, optional
        Duration in seconds of the animation (default is 0.75 seconds)
    fillpercentage : int, optional (in 0,1,2,3,4)
        Amount of the circle that is filled by the slices (default is 2 which means 60%)
    backcolor : str, optional
        Background color of the pie (default is '#f1f1f1')
    dimension : int or float or str, optional
        Side of the drawing. If an integer or a float is passed, the size is intended in pixels units, otherwise a string containing the units must be passed (example: '4vh'). Th default is 400.0 for 400 pixels
    textcolor : str, optional
        Color of text
    fontsize : int, optional
        Dimension of text in pixels (default is 15)
    textweight : int, optional
        Weight of text (default is 400, >= 500 is Bold)
    decimals : int, optional
        Number of decimal to use for the display of numbers (default is 1)
    centertext : str, optional
        Text string to display at the center of the pie (default is '')
    centercolor : str, optional
        Color to use for the central text
    centerfontsize : int, optional
        Text dimension for the text displayed at the center of the pie
    centertextweight : int, optional
        Weight of central text (default is 500, >= 500 is Bold)
    onclick : function, optional
        Python function to call when the user clicks on one of the slices of the pie. The function will receive as first parameter the index of the clicked slice, and the additional_argument as second parameter
    additional_argument : any, optional
        Additional parameter passed to the onclick function when the user clicks on one of the slices of the pie (default is None)
    is_selected : bool, optional
        Flag to select the pie chart (default is False)
    displayvalues: bool, optional
        If True each slide of the pie will display, inside parenthesis, the corresponding value (default is True)
   
    
    Return
    ------
        a tuple containing an instance of ipywidgets.Output() widget, and a string containing the SVG code of the drawing
    
    Example
    -------
    Example of a pie chart::
    
        from vois import svgUtils
        import plotly.express as px
        from ipywidgets import widgets

        debug = widgets.Output()
        display(debug)

        def onclick(arg):
            with debug:
                print('clicked %s' % arg)

        out, txt = svgUtils.AnimatedPieChart(values=[10.0, 25.0, 18.0, 20.0, 9.5],
                                             labels=['Option<br>1', 'Option<br>2', 
                                                     'Option 3', 'Option 4',
                                                     'Others'], 
                                             centerfontsize=28,
                                             fontsize=16, textweight=400,
                                             colors=px.colors.qualitative.D3,
                                             backcolor='#dfdfdf',
                                             centertext='Example Pie',
                                             onclick=onclick,
                                             dimension=380.0,
                                             duration=1.0)
        display(out)    


    .. figure:: figures/pieChart2.png
       :scale: 100 %
       :alt: pieChart example

       Example of an animated SVG to graphically represent a pie chart
    """
    
    

    # From polar to cartesian coordinates
    def polar2cart(r, phi):
        arad = math.radians(-phi)
        x = r * math.cos(arad)
        y = r * math.sin(arad)
        return x,-y

    # From cartesian to polar coordinates: returns r, theta(degrees)`
    def cart2polar(x,y):
        return math.hypot(x,y), math.degrees(math.atan2(y,x))

    # Calculate text position: returns x,y
    def textPosition(angle, r):
        while angle > 360: angle -= 360
        while angle < 0: angle += 360
        return polar2cart(r, angle)

        
    if fillpercentage == FillPercentage.fill100:
        textdistance = 120
        width = 100.0
    elif fillpercentage == FillPercentage.fill75:
        textdistance = 125
        width = 75.0
    elif fillpercentage == FillPercentage.fill60:
        textdistance = 140
        width = 60.0
    elif fillpercentage == FillPercentage.fill50:
        textdistance = 150
        width = 50.0
    else:
        textdistance = 170
        width = 40.0

    svgdimension = 400
    
    # Centre and external radius of the circle
    cx = svgdimension/2
    cy = svgdimension/2
    externalradius = svgdimension/2
    
    # Internal Radius
    r = 200.0 - width
    stroke = width * 2.0 - 10.0
    
    
    
    # Percentages calculation
    tot = sum(values)
    perc = [100.0*x/tot for x in values]
        

    # Pre-processing of all labels
    if not labels is None:
        for i in range(len(labels)):
            labels[i] = labels[i].replace('-',' ').replace('_',' ')
    
    
    # Calculation of all coordinates and dimensioning: key is the index of the sector, from 0 to len(values)-1
    coordinates = {}

    startangle = -90.0
    for i in range(len(perc)):
        angle = 360.0 * perc[i] / 100.0

        # Sorry!
        if fillpercentage == FillPercentage.fill100:
            span  = 628.8 * perc[i] / 100.0
        elif fillpercentage == FillPercentage.fill75:
            span  = 790.0 * perc[i] / 100.0
        elif fillpercentage == FillPercentage.fill60:
            span  = 890.0 * perc[i] / 100.0
        elif fillpercentage == FillPercentage.fill50:
            span  = 945.0 * perc[i] / 100.0
        else:
            span  = 1010.0 * perc[i] / 100.0
                
        color = colors[i % len(colors)]
        color_no_answer = '#757575'

        strvalue = '{:.{prec}f}'.format(perc[i], prec=decimals)

        if not labels is None and i < len(labels):
            if displayvalues:
                text = '%s<br>%s%%<br>(%s)' % (labels[i],strvalue,values[i])
            else:
                text = '%s<br>%s%%' % (labels[i],strvalue)
                     
            
            if text[:9] == 'no answer' or text[:12] == 'no<br>answer': color = color_no_answer    # Request for same color for all the "no answer" slices
            #print(text)
        else:
            if displayvalues:
                text = strvalue + '<br>(' + str(values[i])+ ')'
            else:
                text = strvalue
        
        fulltext = centertext + ': ' + text.replace('<br>',' ')
        tooltip = '<title>%s</title>' % (fulltext)
            
        x,y = textPosition(startangle+angle/2.0, textdistance)

        svgtext = ''
        if perc[i] > 5.0:

            vtext = text.split('<br>')
            y -= fontsize * (len(vtext) - 1.0)

            svgtext = '<text id="text" class="portion-text" style="pointer-events: none" font-size="%f" text-anchor="middle" x="%f" y="%f" fill="%s" font-weight="%f" style="font-family: %s;">' % (fontsize, x+200,y+225, textcolor, textweight, fontsettings.font_name)

            for t in vtext:
                svgtext += '<tspan x="%f" dy="%f">%s</tspan>' % (x+200, fontsize, t)

            svgtext += '</text>'

                
        coordinates[i] = {
                          'startangle': startangle,
                          'angle'     : angle,
                          'span'      : span,
                          'color'     : color,
                          'text'      : text,
                          'tooltip'   : tooltip,
                          'x'         : x,
                          'y'         : y,
                          'svgtext'   : svgtext
                         }

        startangle += angle


    underline = ''
    if is_selected:
        class_name       = 'portion-out'
        underline        = 'text-decoration="underline"'
        centertextweight = 700
        #centercolor      = settings.color_first
    else:
        class_name = 'portion'
        
        
        
    # Creation of the SVG
    def createSVG():
        svg = '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" x="0px" y="0px" viewBox="0 0 %d %d" xml:space="preserve">' % (svgdimension,svgdimension)

        # CSS styling
        svg += '''
            <style type="text/css">
               @import url('%s');
               .portion:hover , .portion:hover + .portion-text {
                    cursor: pointer;
                    stroke-width: %f;
                    font-weight: %d;
                    text-decoration: underline;
                    }
            </style>
        ''' % (fontsettings.font_url, stroke*1.1, textweight+0)


        # Back circle
        strclass  = ''
        if is_selected:
            strclass = 'class="portion-out"'
        svg += '<circle %s fill="%s" cx="200" cy="200" r="200" />' % ( strclass, backcolor )
        
        if is_selected:
            svg += '<circle %s fill="%s" cx="200" cy="200" r="85" />' % ( strclass, settings.color_first )


        # Animated circles
        svgtextall = ''
            
        for i in range(len(perc)):
            angle      = coordinates[i]['angle']
            startangle = coordinates[i]['startangle']
            span       = coordinates[i]['span']

            color   = coordinates[i]['color']
            text    = coordinates[i]['text']
            tooltip = coordinates[i]['tooltip']

            x = coordinates[i]['x']
            y = coordinates[i]['y']

            # Text
            svgtext = coordinates[i]['svgtext']
            svgtextall += svgtext

            # Circle sector
            if duration <= 0.0:
                svg += '''
                        <circle class="%s" cx="%d" cy="%d" r="%f" transform="rotate(%f, 200, 200)" stroke-dasharray="%d, 1000" fill="none" stroke-width="%f" stroke="%s" stroke-linecap="butt">%s</circle>
                            %s
                ''' % (class_name, cx,cy, r, startangle, int(span), stroke, color, tooltip, svgtext)
            else:
                svg += '''
                        <circle class="%s" cx="%d" cy="%d" r="%f" transform="rotate(%f, 200, 200)" stroke-dasharray="0, 1000" fill="none" stroke-width="%f" stroke="%s" stroke-linecap="butt">
                            <animate attributeName="stroke-dasharray" dur="%fs" to="%d,1000" fill="freeze" />%s</circle>
                            %s
                ''' % (class_name, cx,cy, r, startangle, stroke, color, duration, int(span), tooltip, svgtext)

        # Center text
        svg += '<text style="pointer-events: none" font-size="%f" text-anchor="middle" x="200" y="%d" fill="%s" font-weight="%f" %s style="font-family: %s;">%s</text>' % (centerfontsize, 200+fontsize/4, centercolor, centertextweight, underline, fontsettings.font_name, centertext)
                

        # Display text again so that it is always visible!!!
        svg += svgtextall

        svg += '</svg>'
        return svg
    
    
    # Create an output widget and display SVG in it
    if isinstance(dimension, int):
        w = '%dpx' % dimension
    elif isinstance(dimension, float):
        w = '%fpx' % dimension
    else:
        w = str(dimension)
    out = widgets.Output(layout=Layout(width=w, height='calc(%s + 14px)' % w))

    
    svg = createSVG()
    with out:
        display(HTML(svg))
        
    # Add an event manager to the out Output widgets
    d = Event(source=out, watched_events=['click'])

    cx = svgdimension/2.0
    cy = svgdimension/2.0
    radius = svgdimension/2.0
    
    if fillpercentage == FillPercentage.fill100:
        rmin = 0.0
    elif fillpercentage == FillPercentage.fill75:
        rmin = radius * 0.25
    elif fillpercentage == FillPercentage.fill60:
        rmin = radius * 0.40
    elif fillpercentage == FillPercentage.fill50:
        rmin = radius * 0.50
    else:
        rmin = 1.0

    
    # Convert clik angle to startangle values
    def clickangle2startangle(a):
        # Convert in [0,360)
        if a < 0:
            a += 360.0
            
        # Invert direction
        sa = -a
        
        # Take inside interval [-90,270)
        if sa < -90.0:
            sa += 360.0
            
        return sa
        
        
    def handle_event(event):
            
        if is_selected:

            # Call callback function
            if not onclick is None:
                if not additional_argument is None:
                    onclick(-1, additional_argument)
                else:
                    onclick(-1)
                
            out.clear_output(wait=True)
            with out:
                display(HTML(createSVG()))

        else:
            x = event['relativeX']
            y = event['relativeY']
            w = event['boundingRectWidth']
            h = event['boundingRectHeight']
            xp = (x / w) * svgdimension
            yp = (y / h) * svgdimension
            px = xp - cx
            py = cy - yp
            
            #with out:
            #    print(x,y, xp,yp )
                
            clickdist,clickangle = cart2polar(px,py)
            if clickdist >= rmin and clickdist <= radius:
                sa = clickangle2startangle(clickangle)
                #with out:
                #    print(clickangle, sa)
                    
                for i in range(len(perc)):
                    angle      = coordinates[i]['angle']
                    startangle = coordinates[i]['startangle']
                
                    if sa >= startangle and sa <= startangle+angle:

                        # Call callback function
                        if not onclick is None:
                            if not additional_argument is None:
                                onclick(i, additional_argument)
                            else:
                                onclick(i)
                            
                        #out.clear_output(wait=True)
                        #with out:
                        #    display(HTML(createSVG()))
                            
                        break
       
    d.on_dom_event(handle_event)
    return out, svg




###########################################################################################################################################################################
# Title
###########################################################################################################################################################################
def svgTitle(title='Dashboard title', subtitle1='Subtitle1', subtitle2='Subtitle2', xline=290):
    """
    Creation of a simple title for a dashboard in SVG format
    
    Parameters
    ----------
    title : str, optional
        Title of the dashboard (default is 'Dashboard title')
    subtitle1 : str, optional
        First line of the subtitle (default is 'Subtitle1')
    subtitle2 : str, optional
        Second line of the subtitle (default is 'Subtitle2')
    xline : int, optional
        X position of a vertical line to divide title from subtitles (default is 290)
        
    Returns
    -------
        a string containing a SVG drawing that spans 100% of the width and 46 pixels in height
        
    Example
    -------
    Example of a title and logo SVG::
    
        from vois import svgUtils
        from ipywidgets import HTML, widgets, Layout

        outTitle = widgets.Output(layout=Layout(width='99%',                   height='64px'))
        outLogo  = widgets.Output(layout=Layout(width='1%', min_width='110px', height='82px'))

        outTitle.clear_output()
        with outTitle:
            display(HTML(svgUtils.svgTitle()))

        outLogo.clear_output()
        with outLogo:
            display(HTML(svgUtils.svgLogo()))

        display(widgets.HBox([outTitle,outLogo]))
        
    .. figure:: figures/svgTitle.png
       :scale: 100 %
       :alt: svgTitle example

       Example of a simple svgTitle and svgLogo
        
    Note
    ----
    This function is completely superseeded by the more complete examples of dashboard titles available using the :py:class:`title.title` or :py:class:`app.app`
    
    """
    svg = '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" x="0px" y="0px" width="100%" height="46" xml:space="preserve">'

    LastData = datetime.today().strftime('%d %b %Y')

    svg += '<text x="0" y="36" font-size="40" fill="%s" font-weight="400">%s</text>' % (settings.label_textcolor,title)
    
    svg += '<line x1="%d" y1="0" x2="%d" y2="38" style="stroke:%s;stroke-width:1.0" />' % (xline, xline, settings.line_color)
    
    svg += '<text x="%d" y="20" font-size="18" fill="%s" font-weight="400">%s</text>' % (xline+15, settings.label_textcolor, subtitle1)
    svg += '<text x="%d" y="38" font-size="18" fill="%s" font-weight="400">%s</text>' % (xline+15, settings.label_textcolor, subtitle2)

    svg += '<text x="60%%" y="24" text-anchor="middle" font-size="26" fill="%s" font-weight="400">%s</text>' % (settings.label_textcolor, LastData)
    svg += '<text x="60%%" y="40" text-anchor="middle" font-size="12" fill="%s" font-weight="400">Last data update</text>' % (settings.label_textcolor)

    svg += '<line x1="0" y1="45" x2="100%%" y2="45" style="stroke:%s;stroke-width:0.3" />' % (settings.line_color)

    svg += '</svg>'
    return svg




###########################################################################################################################################################################
# Logo
###########################################################################################################################################################################
def svgLogo():
    """
    Creation of a simple SVG to display the logo of the European Commission
    
    Returns
    -------
        a string containing a SVG drawing that spans for 68 pixels in width and 44 pixels in height
        
    Example
    -------
    See example provided for the :py:func:`svgUtils.svgTitle` function

    Note
    ----
    This function is completely superseeded by the more complete examples of dashboard titles available using the :py:class:`title.title` or :py:class:`app.app`
    
    """
    svg = '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" x="0px" y="0px" width="68px" height="44px" xml:space="preserve">'
    svg += '<image href="https://jeodpp.jrc.ec.europa.eu/services/shared/Notebooks/images/European_Commission.svg" x="0" y="1" width="68" height="44" />'
    svg += '</svg>'
    return svg




"""Calendar widget showing days with events"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
from ipywidgets import widgets, Layout

import collections
from datetime import datetime, date, timedelta
from dateutil.relativedelta import relativedelta

try:
    from . import settings
except:
    import settings

    
    
# Given two dates as strings in YYY-MM-DD format, returns the number of weeks of covering the period between start and end day
def number_of_weeks(start, end):
    start_date = datetime.strptime(start, '%Y-%m-%d').date()
    end_date   = datetime.strptime(end,   '%Y-%m-%d').date()
    start_monday = start_date + timedelta(days=-start_date.weekday())
    end_monday   = end_date   + timedelta(days=-end_date.weekday())
    delta = end_monday - start_monday
    return 1 + delta.days // 7

#####################################################################################################################################################
# Calendar that displays days inside an interval of dates and optional events
#####################################################################################################################################################
class dayCalendar():
    """
    Input widget to display a daily calendar for a range of dates and allows for highlighting some of the days and manages the click on the days.

    Parameters
    ----------
    start : str or datetime or date instance, optional
        Initial date of the calendar as a string in format 'YYYY-MM-DD' or as an instance of datetime.datetime or datetime.date (default is today date minus one month)
    end : str or datetime or date instance, optional
        Final date of the calendar as a string in format 'YYYY-MM-DD' or as an instance of datetime.datetime or datetime.date  (default is today)
    color : str, optional
        Color to use for the highlighting of days in the calendar (default is settings.color_first)
    dark : bool, optional
        If True, the calendar will have a dark background (default is settings.dark_mode)
    days: list of str, optional
        List of days to be highlighted as strings in "YYYY-MM-DD" format (default is []). The list can contain repeated days (see show_count below).
    show_count: bool, optional
        If True, the event bar will show the number of events on each of the highlighted days (default is False)
    width : int, optional
        Width of the widget in pixels (default is 340)
    height : int, optional
        Height of the widget in pixels (default is None). If None is passed, the height will be calculated depending on the range of dates defined by start and end parameters.
    on_click : function, optional
        Python function to call when the user clicks on one day of the calendar. The function will receive as parameter a string in "YYYY-MM-DD" format. (default is None)
    on_click _event: function, optional
        Python function to call when the user clicks on the highlighting bar of one day of the calendar. The function will receive as parameter a string in "YYYY-MM-DD" format. (default is None)

    Example
    -------
    Creation of a date picker widget::
        
        from vois.vuetify import dayCalendar
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def on_click(day):
            with output:
                print('Clicked on ', day)

        c = dayCalendar.dayCalendar(start='2023-10-01', end='2023-10-31',
                                    days=['2023-10-10', '2023-10-20'],
                                    on_click=on_click)

        display(c.draw())
        display(output)

    .. figure:: figures/dayCalendar.png
       :scale: 100 %
       :alt: dayCalendar widget

       Example of a dayCalendar
    """
    
    
    # Initialization
    def __init__(self,
                 start=date.today() + relativedelta(months=-1), # Dates as strings in "YYYY-MM-DD" format or datetime or date instances
                 end=date.today(),
                 color=settings.color_first,                    # Color used to highlight the events
                 dark=settings.dark_mode,
                 days=[],                                       # List of strings in "YYYY-MM-DD" format to highlight some of the days
                 show_count=False,                              # If True shows in each higlighted day one char '°' for each repetition inside the days list
                 width=340,                                     # Width on pixels
                 height=None,                                   # Height in pixels
                 on_click=None,                                 # Function called at the click on a day (will receive the day as string in "YYYY-MM-DD" format as parameter)
                 on_click_event=None                            # Function called at the click on an event (will receive the day as string in "YYYY-MM-DD" format as parameter)
                ):
        
        if isinstance(start,datetime) or isinstance(start,date):
            self.start = start.strftime('%Y-%m-%d')
        else:
            self.start = start

        if isinstance(end,datetime) or isinstance(end,date):
            self.end = end.strftime('%Y-%m-%d')
        else:
            self.end = end
            
        self._color         = color
        self._days          = days
        self.show_count     = show_count
        self.width          = width
        self.height         = height
        self.on_click       = on_click
        self.on_click_event = on_click_event
        
        self.cal = v.Calendar(v_model='', start=self.start, end=self.end, now='1899-12-31', type='custom-weekly',
                              event_more=False, event_height=6, events=[], event_color=self._color, short_weekdays=True,
                              hide_header=False, show_month_on_first=True, weekdays=[1,2,3,4,5,6,0], dark=dark)                  # Week start on Monday: standard ISO!
        self.cal.on_event('input', self.__internal_on_click)
        self.cal.on_event('click:event', self.__internal_on_click_event)

        self.days2events()
        
        card_height = 53 * number_of_weeks(self.start, self.end)
        self.card = v.Card(flat=True, children=[self.cal], width=self.width, height=card_height, class_='pa-0 ma-0 mb-1')

        if self.height is None:
            self.height = 10 + card_height
            
        self.output = widgets.Output(layout=Layout(height='%dpx'%self.height))
        with self.output:
            display(self.card)
         
    
    # Convert a list of days in events for the calendar widget
    def days2events(self):
        if self.show_count:
             # See https://en.wikipedia.org/wiki/List_of_Unicode_characters
            self.events = [{'name': '\u02DA'*count, 'start': day } for day,count in collections.Counter(self._days).items()]
        else:
            self.events = [{'name': '', 'start': d} for d in list(set(self._days))]
        self.cal.events = self.events

        
    # Manage click on a day
    def __internal_on_click(self, widget, event, data):
        if not self.on_click is None:
            self.on_click(data)
        
        
    # Manage click on an event
    def __internal_on_click_event(self, widget, event, data):
        if 'event' in data:
            day = data['event']['start']
            if not self.on_click_event is None:
                self.on_click_event(day)
            
            
    # Returns the vuetify object to display (the Output widget containing the card containing the calendar)
    def draw(self):
        """Returns the ipyvuetify object to display (the internal Output widget)"""
        return self.output

    
    # color property
    @property
    def color(self):
        """
        Get/Set the color of the highlighted days
        
        Returns
        --------
        color : str
            Color of the highlighted days in the calendar

        Example
        -------
        Programmatically change the color::
            
            cal.color = 'red'
            print(cal.color)
        
        """
        return self._color
    
    @color.setter
    def color(self, col):
        self._color = col
        self.cal.event_color = self._color
    
    
    # days property
    @property
    def days(self):
        """
        Get/Set the highlighted days
        
        Returns
        --------
        listfodays : list of strings in "YYYY-MM-DD" format
            List of days currently highlighted in the calendar

        Example
        -------
        Programmatically change the highlighted days::
            
            cal.days = ['2023-10-15', '2023-10-25']
            print(cal.days)
        
        """
        return self._days
    
    @days.setter
    def days(self, listofdays):
        self._days = listofdays
        self.days2events()


"""Cards with title, subtitle and image displayed in rows and columns"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
import traitlets


class cardsGrid(v.VuetifyTemplate):
    """
    Cards with title, subtitle and image displayed in a grid of rows and columns.
        
    Parameters
    ----------
    cards : list of json element, one for each card to display, optional
        Each of the json elements must have this structure: { "title": "", "subtitle": "", "image": ""}, optional tags are "color", "imagesize", "icon" and "iconsize", "titletooltip"
    width : str, optional
        Width of the cards (default is '400px')
    height : str, optional
        Heigth of the cards (default is '' which means that each card has its own height defined by its content)
    cols : int, optional
        Horizontal column span [1,12] for each of the card (default is 6)
    color : str, optional
        Background color of the cards (default is 'white')
    dark : bool, optional
        If True the title and subtitle texts are displayed in white color, if False they are displayed in black (default is False)
    ripple : bool, optional
        If True the click on the card is highlighted (default is False)
    iconsize : str, optional
        Size of the area where the icon is displayed (default is '32px')
    imagesize : str, optional
        Size of the area where the image is displayed (default is '190px')
    on_click : function, optional
        Python function to call when the user clicks on the card. The function will receive as parameter the index of the clicked card. (default is None)
    responsive : bool, optional
        If True, the font size is automatically changed according to the page size (default is False)
    fontsizemultiplier : float, optional
        Multiply factor for changing the standard size of the font used for title and subtitle (default is 1.0)
    tooltipwidth : str, optional
        Max width of the tooltip window to open on hover on the cards title (default is '600px')

    Example
    -------
    Creation of a cards grid to display text and an image::
        
        from vois.vuetify import cardsGrid
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def on_click(index):
            with output:
                print(index)

        cards = [
           { "title": "title0", "subtitle": "subtitle0", "image": "https://cdn.vuetifyjs.com/images/cards/sunshine.jpg", "titletooltip": "Example of card title tooltip" },
           { "title": "title1", "subtitle": "subtitle1", "image": "https://cdn.vuetifyjs.com/images/cards/road.jpg" },
           { "title": "title2", "subtitle": "subtitle2", "image": "https://cdn.vuetifyjs.com/images/cards/plane.jpg" },
           { "title": "title3", "subtitle": "subtitle3", "image": "https://cdn.vuetifyjs.com/images/cards/house.jpg" }
        ]

        g = cardsGrid.cardsGrid(cards=cards, cols=6, width='350px', imagesize='200px', on_click=on_click)

        display(g)
        display(output)

    .. figure:: figures/cardsGrid.png
       :scale: 100 %
       :alt: cardsGrid widget

       Example of a cardsGrid to display multiple cards containing texts and an images
    """
    
    cards     = traitlets.Any([]).tag(sync=True)           # Cards to display (array of json objects containing "title", "subtitle", "image" tags)
    
    width     = traitlets.Unicode('400px').tag(sync=True)  # Width of the cards
    height    = traitlets.Unicode('').tag(sync=True)       # Height of the cards
    color     = traitlets.Unicode('white').tag(sync=True)  # Background color
    cols      = traitlets.Int(6).tag(sync=True)            # Horizontal column span [1,12] for each of the card
    dark      = traitlets.Bool(False).tag(sync=True)       # Dark mode flag (if True the text is displayed in white)
    ripple    = traitlets.Bool(False).tag(sync=True)       # Ripple flag (if True the click on the card is highlighted)
    on_click  = traitlets.Any(None).tag(sync=False)        # Name of a python function to call when one of the cards is clicked (it will receive as argument the index of the clicked card)
    responsive         = traitlets.Bool(False).tag(sync=True)
    fontsizemultiplier = traitlets.Float(1.0).tag(sync=True)
    tooltipwidth = traitlets.Unicode('600px').tag(sync=True)  # Width of the tooltip


    @traitlets.default('template')
    def _template(self):
        return '''
<v-container fluid>
  <v-row dense>
    <v-col
      v-for="card,index in cards"
      :key="card.title"
      :cols="cols"
    >
        <v-card
            class="pa-0 ma-1"
            :width="width"
            :color="color"
            :dark="dark"
            :height="height"
            hover
            :ripple="ripple"
            link
            @click="clicked(index)"
        >
            <div class="d-flex flex-no-wrap justify-space-between">
              <div>
                  <v-row dense>
                      <v-avatar class="pa-0 ma-0 ml-3 mt-2 mr-n2" :size="card.iconsize" tile >
                          <v-img :src="card.icon"></v-img>
                      </v-avatar>
                      <div :style="card.color">
                         <v-tooltip v-if="card.titletooltip" bottom :max-width="tooltipwidth">
                            <template v-slot:activator="{ on, attrs }">
                               <v-card-title v-bind="attrs" v-on="on" class="mt-n2 mb-1" :style="fontSizeTitle" v-text="card.title"></v-card-title>
                            </template>
                            <span>{{card.titletooltip}}</span>
                         </v-tooltip>
                         <v-card-title v-else class="mt-n2 mb-1" :style="fontSizeTitle" v-text="card.title"></v-card-title>
                      </div>
                  </v-row>
                  <div :style="card.color">
                    <div :class="card.margins" :style="fontSizeSubTitle" v-html="card.subtitle"/>
                  </div>
              </div>
              <v-avatar class="ma-0" :size="card.imagesize" tile >
                  <v-img :src="card.image" contain ></v-img>
              </v-avatar>
            </div>
        </v-card>
    </v-col>
  </v-row>
</v-container>

<script>
  export default {
    computed: {
      fontSizeTitle (num)
        {
         if( this.responsive )
           {
            switch( this.$vuetify.breakpoint.name )
              {
               case 'xs': return 'font-size: ' + (0.75*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'sm': return 'font-size: ' + (1.00*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'md': return 'font-size: ' + (1.25*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'lg': return 'font-size: ' + (1.45*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'xl': return 'font-size: ' + (1.60*this.fontsizemultiplier).toFixed(3) + 'em;'
              }
           }
         else
           {
            return 'font-size: ' + (1.25*this.fontsizemultiplier).toFixed(3) + 'em;'
           }
      },
        
      fontSizeSubTitle ()
        {
         if( this.responsive )
           {
            switch( this.$vuetify.breakpoint.name )
              {
               case 'xs': return 'font-size: ' + (0.8*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'sm': return 'font-size: ' + (0.9*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'md': return 'font-size: ' + (1.0*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'lg': return 'font-size: ' + (1.1*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'xl': return 'font-size: ' + (1.2*this.fontsizemultiplier).toFixed(3) + 'em;'
              }
           }
         else
           {
            return 'font-size: ' + (1.0*this.fontsizemultiplier).toFixed(3) + 'em;'
           }
        },
    },
  }
</script>
'''

    def __init__(self, *args,
                 cards=[],
                 width='400px',
                 height='',
                 color='white',
                 cols=6,
                 dark=False,
                 ripple=False,
                 iconsize='32px',
                 imagesize='190px',
                 on_click=None,
                 responsive=False,
                 fontsizemultiplier=1.0,
                 tooltipwidth='600px',
                 **kwargs):
        
        for c in cards:
            if not "color" in c:
                if dark:
                    c["color"] = "color: white;"
                else:
                    c["color"] = "color: black;"
            else:
                c["color"] = "color: %s;" % c["color"]

            if not "iconsize" in c:
                c["iconsize"] = iconsize

            if not "icon" in c:
                c["icon"] = ""
                c["iconsize"] = '0px'

            if not "imagesize" in c:
                c["imagesize"] = imagesize

            if not "subtitle" in c or len(c["subtitle"]) == 0:
                c["margins"] = "ma-0 ml-4 mb-n3"
            else:
                c["margins"] = "ma-0 ml-4 mb-3"
                
        self.cards  = cards
        self.width  = width
        self.height = height
        self.color  = color
        self.cols   = cols
        self.dark   = dark
        self.ripple = ripple
        self.on_click = on_click
        
        self.responsive         = responsive
        self.fontsizemultiplier = fontsizemultiplier
        self.tooltipwidth = tooltipwidth
          
        super().__init__(*args, **kwargs)
        
    
    # Manage "click" event
    def vue_clicked(self, data):
        if not self.on_click is None:
            self.on_click(data)


"""Class that implements a title bar that can be used as a simple main interface for a dashboard."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
    from . import fontsettings
except:
    import settings
    import fontsettings


#####################################################################################################################################################
# Display a title panel with some buttons
#####################################################################################################################################################
class title():
    """
    Class that implements a title bar that can be used as a simple main interface for a dashboard.
        
    Parameters
    ----------
    text : str, optional
        Text to display in the title widget
    textweight : int, optional
        Weight of the title text (default is 500)
    height : int, optional
        Height of the title bar in pixels (default is 85 pixels)
    color : str, optional
        Background color of the title bar (default is the color_first defined in the settings.py module)
    dark : bool, optional
        Flag that controls the color of the text in foreground (if True, the text will be displayed in white, elsewhere in black)
    buttons : list of strings, optional
        List of strings to be used as text of the buttons to display below the title bar main text
    onclick : function, optional
        Python function to call when the user clicks on one of the buttons. The function will receive a parameter of string containing the name of the button clicked 
    menu : bool, optional
        Flag that controls the display of a menu icon on the left side of the title bar (default is True)
    menumarginy : int, optional
        Vertical displace of the menu icon from the top of the title bar (default is 2)
    onmenuclick :
        Python function to call when the user clicks on the menu icon. The function will receive no parameters
    logo : str, optional
        String conaining the URL of the logo image to display on the right side of the title bar
    logowidth : int, optional
        Width in pixels of the area where the logo has to be displayed
    logomarginy : int, optional
        Vertical displace of the logo from the top of the title bar (default is 0)
    output : ipywidgets.Output, optional
        Output widget on which the title bar has to be displayed

    Example
    -------
    Creation and display of a title bar with additional buttons and left menu icon::
        
        from vois.vuetify import title
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()
        display(output)

        def onclick(arg):
            with output:
                print(arg)

        def onmenu():
            with output:
                print('MENU')

        f = title.title(text='Title text to display', color='amber',
                        menu=True, onmenuclick=onmenu,
                        buttons=['Home', 'About Us', 'Team', 'Services', 'Blog', 'Contact Us'],
                        logo='https://jeodpp.jrc.ec.europa.eu/services/shared/home/images/JRCBigDataPlatform2.png',
                        logomarginy=6, menumarginy=4,
                        height=80, onclick=onclick, output=output)

    .. figure:: figures/title.png
       :scale: 100 %
       :alt: title widget

       Example of a title bar.
    """
    
    def __init__(self, text='', textweight=500, buttons=[], menu=True, onmenuclick=None, dark=settings.dark_mode, height=85, color=settings.color_first, onclick=None,
                 logo='https://jeodpp.jrc.ec.europa.eu/services/shared/Notebooks/images/European_Commission.svg', logowidth=100, logomarginy=0, menumarginy=2,
                 output=None):
        
        self.onclick     = onclick
        self.onmenuclick = onmenuclick
        
        if dark:
            textcol = settings.textcolor_dark
        else:
            textcol = settings.textcolor_notdark
        
        s = 'font-family: %s; font-size: 30px; font-weight: %d; text-transform: none; overflow: hidden;' % (fontsettings.font_name, textweight)
        h = v.Html(tag='div', style_=s, children=[text])
        
        c = v.Col(color=color, class_="pa-0 ma-0 mt-n1 mb-n2 text-center %s--text" % textcol, cols="12", children=[h])
    
        self.buttons_text = buttons
        self.buttons = [v.Btn(color=textcol, text=True, rounded=True, class_='my-2', children=[x]) for x in self.buttons_text]
        for b in self.buttons:
            b.on_event('click', self.__internal_onclick)
            
        if menu:
            m = v.Btn(icon=True, dark=dark, children=[v.Icon(children=['mdi-menu'])])
            m.on_event('click', self.__internal_onmenu)
        else:
            m = v.Spacer()

        r = v.Row(class_='pa-0 ma-0 mb-n2', justify="center", no_gutters=True, children=self.buttons, style_='overflow: hidden;')

        img = v.Img(src=logo, class_='pa-0 ma-0 mt-%d mr-1'%logomarginy, max_width=logowidth)
        c0 = v.Col(cols="1",  children=[m], class_='pa-0 ma-0 ml-1 mr-n1 mt-%d' % menumarginy)
        c1 = v.Col(cols="10", children=[c,r])
        c2 = v.Col(cols="1",  children=[img], class_='pa-0 ma-0 mt-1')
        rfull = v.Row(class_='pa-0 ma-0', no_gutters=True, children=[c0,c1,c2], style_='overflow: hidden;')
        
        self.f = v.Footer(color=color, padless=True, children=[rfull], style_='overflow: hidden;')
        if not height is None:
            self.f.height = height
        
        # Display of the footer
        if not output is None:
            with output:
                display(self.f)
        
    # Manage click on the buttons
    def __internal_onclick(self,widget,event,data):
        i = self.buttons.index(widget)
        if not self.onclick is None:
            self.onclick(self.buttons_text[i])

    # Manage click on the menu
    def __internal_onmenu(self, *args):
        if not self.onmenuclick is None:
            self.onmenuclick()
            


"""Circular progress bar to use for lenghty operations."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v
from vois.vuetify.utils.util import *


#####################################################################################################################################################
# Display a circular progress bar centered inside an output widget
#####################################################################################################################################################
class Progress:
    """
    Circular progress bar to use for lenghty operations.
        
    Parameters
    ----------
    output : ipywidgets.Output
        Output widget on which the progress bar has to be displayed
    text : str, optional
        Small text to display inside the circle (default is '')
    show : bool, optional
        If True, display the progress upon creation (default is False)
    size : int, optional
        Diameter of the circle in pixels (default is 120 pixels)
    width : int, optional
        Width in pixels of the moving line (default is 15)
    output_height : int, optional
        Height in pixels of the output widget (default is 400 pixels)
    onchange : function, optional
        Python function to call when the user selects one of the values in the list. The function will receive no parameter (use value property to retrieve the current selection)
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
            
    Example
    -------
    Creation and display of a progress widget::
        
        from vois.vuetify import progress
        from ipywidgets import widgets, Layout
        from IPython.display import display

        output_height = 500
        output = widgets.Output(layout=Layout(width='99%', height='%dpx' % (output_height+10)))
        display(output)

        p = Progress(output, text='Please, wait...',
                              size=200, width=20,
                              show=True, output_height=output_height)
        
    To close the progress::
    
        p.close()
        

    .. figure:: figures/progress.png
       :scale: 100 %
       :alt: progress widget

       Example of progress widget displayed inside an Output.
   """
    deprecation_alias = dict(outputheight='output_height')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 output,  # Output widget where the progress must be displayed
                 text: str = '',  # Small text inside the circle
                 show: bool = False,  # If True, display the progress upon creation
                 size: int = 120,  # Diameter of the circle in pixels
                 width: int = 15,  # Width of the moving line
                 output_height: int = 400,  # Height in pixels of the output widget
                 color: str = None):  # Color

        from vois.vuetify import settings

        self.output = output
        self.size = size
        self.output_height = output_height

        self.prog = v.ProgressCircular(style_='overflow: hidden;',
                                       class_='pa-0 ma-0',
                                       indeterminate=True,
                                       size=size,
                                       width=width,
                                       children=[text],
                                       color=color if color is not None else settings.color_first)

        self.nav = None

        if show:
            self.show()

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

    # Display the progress
    def show(self):
        """Displays the progress into the output widget"""
        if self.output:
            self.output.clear_output(wait=True)
            with self.output:
                # Centered vertically and horizontally!
                h = v.Html(tag='div',
                           style_="height: %dpx; display: flex; justify-content: center; align-content: center; align-items:center; flex-direction: column; " % self.output_height,
                           class_="text-center", children=[self.prog])

                display(h)

    # Display the progress as an overlay in a specific position
    def showAbsolute(self, output, left, top, zindex=9999):
        """Displays the progress as an overlay layer in a specific position
        
        Parameters
        ----------
        output : ipywidgets.Output
            Output widget to use for the display of the transparent widget
        left : str
            Position of the left side of the widget. It can be in pixels, vw, or other units.
        top : str
            Position of the top side of the widget. It can be in pixels, vh, or other units.
        zindex : int, optional
            Z-index of the progress on the page (default is 9999)
        """
        self.nav = v.NavigationDrawer(stateless=True, permanent=True, floating=True, fixed=True, left=True,
                                      color="transparent",
                                      width="%dpx" % self.size,
                                      height="%dpx" % self.size,
                                      style_="left: %s; top: %s; z-index: %d;" % (left, top, zindex),
                                      class_="pa-0 ma-0", children=[self.prog])
        with output:
            display(self.nav)

    # Hide the progress
    def close(self):
        """Hides the progress by clearing the output widget content"""
        if self.output:
            self.output.clear_output()

        if self.nav:
            self.nav.close()



from warnings import warn
from vois.vuetify import Progress
from vois.vuetify.utils.util import deprecation_class_warning


class progress(Progress):

    def __init_subclass__(cls, **kwargs):
        deprecation_class_warning(cls.__name__)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        deprecation_class_warning(self.__class__.__name__)



"""Font settings for ipyvuetify widgets."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Font name and URL
font_name = 'Roboto'
font_url  = 'https://fonts.googleapis.com/css?family=%s:400,100,100italic,300,300italic,400italic,500,500italic,700,700italic,900,900italic' % (font_name)



"""Dialog-box to ask a yes-no question to the user."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
    from . import tooltip
    from . import dialogGeneric
except:
    import settings
    import tooltip
    import dialogGeneric

    
    
#####################################################################################################################################################
# Display a dialog having "yes" and "no" buttons
#####################################################################################################################################################
class dialogYesNo(dialogGeneric.dialogGeneric):
    """
    Dialog-box to ask a yes-no question to the user.
    
    Derived class from dialogGeneric.dialogGeneric.
        
    Parameters
    ----------
    title : str, optional
        Title of the dialog-box to be displayed in the top toolbar (default is '')
    dark : bool, optional
        Flag that controls the color of the text in foreground (if True, the text will be displayed in white, elsewhere in black)
    show : bool, optional
        Flag to immediately show the dialog-box upon creation (default is False)
    width : int or str, optional
        Width of the dialog-box. If an integer is passed the width is intended in pixels. Default is 500 pixels
    addclosebuttons : bool, optional
        If True, the dialog will have a 'close' button in the top toolbar (default is True)
    transition : str, optional
        Transition to use for the dialog display and close (default is 'dialog-fade-transition'. See: https://vuetifyjs.com/en/styles/transitions/ for a list of available transitions (substitute 'v-' with 'dialog-'')
    output : ipywidgets.Output, optional
        Output widget on which the widget has to be displayed
    titleheight : str, optional
        Height of the title toolbar. It can be: 'prominent', 'dense', 'extended' or a value in pixels (default is 'dense')
    on_yes : function, optional
        Python function to call when the user clicks on the YES button. The function will receive no parameters (default is None)
    on_no : function, optional
        Python function to call when the user clicks on the NO button. The function will receive no parameters (default is None)
            
    Example
    -------
    Creation and display of a Yes-No dialog box::
        
        from vois.vuetify import dialogYesNo
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()
        display(output)

        def on_yes():
            with output:
                print('YES')

        def on_no():
            with output:
                print('NO')

        dlg = dialogYesNo.dialogYesNo(title='Question',
                                      text='Confirm removal of the selected file?',
                                      titleheight=40, width=400, output=output,
                                      show=True, transition='dialog-bottom-transition',
                                      on_yes=on_yes, on_no=on_no)
                              
    .. figure:: figures/dialogYesNo.png
       :scale: 100 %
       :alt: dialogYesNo widget

       Example of a dialogYesNo to ask a yes-no question to the user.
   """
        
    def __init__(self, on_yes=None, on_no=None, *args, **kwargs):
        
        self.on_yes = on_yes
        self.on_no  = on_no
        
        # Create yes and no buttons
        byes = v.Btn(text=True, children=['Yes'])
        byes.on_event('click', self.__internal_on_yes)
        byes = tooltip.tooltip('Answer YES and close',byes)

        bno = v.Btn(text=True, children=['No'])
        bno.on_event('click', self.__internal_on_no)
        bno = tooltip.tooltip('Answer NO and close',bno)
        
        r = v.Row(no_gutters=True, justify="end", children=[byes,bno])
        
        # Create the content to pass to the dialogGeneric
        kwargs['content'] = [v.Html(tag='div', class_="pa-0 ma-4 mt-4 mb-2", children=[kwargs['text']]), r]
        kwargs['text'] = ''
        kwargs['fullscreen'] = False
        kwargs['persistent'] = True
        
        super().__init__(*args, **kwargs)

    # Cliked yes
    def __internal_on_yes(self,*args):
        self.dialog.v_model = False
        if not self.on_yes is None:
            self.on_yes()

    # Cliked no
    def __internal_on_no(self,*args):
        self.dialog.v_model = False
        if not self.on_no is None:
            self.on_no()
        


"""Radio buttons to allow users to select from a predefined set of options."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

from vois.vuetify import tooltip
from vois.vuetify.utils.util import *
from typing import Callable, Optional


#####################################################################################################################################################
# Radio control
#####################################################################################################################################################
class Radio(v.RadioGroup):
    """
    Radio buttons to allow users to select from a predefined set of options.
        
    Parameters
    ----------
    index : int
        Index of the option initially selected (from 0 to len(labels)-1)
    labels : list of strings
        Strings to be displayed as options in the radio widget
    tooltips : list of str, optional
        List of strings to use as tooltips for the corresponding radio items (default is None)
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    on_change : function, optional
        Python function to call when the user selects one of the values in the list. The function will receive a single parameter, containing the index of the selected option in the range from 0 to len(labels)-1
    row : bool, optional
        Flag to control the position of radio buttons, either horizontal or vertical (default is True)
            
    Example
    -------
    Creation and display of a radio widget to select among three options::
        
        from vois.vuetify import Radio
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def on_change(value):
            with output:
                print(value)

        r = Radio(0,
                ['Option 0', 'Option 1', 'Option 2'],
                tooltips=['Tooltip for Option 1'],
                on_change=on_change,
                row=True)

        display(r)
        display(output)

    .. figure:: figures/radio.png
       :scale: 100 %
       :alt: radio widget

       Example of a radio widget to select from three options.
   """

    deprecation_alias = dict(onchange='on_change')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 index: int,
                 labels: list[str],
                 tooltips: list[str] = [],
                 color: str = None,
                 on_change: Optional[Callable[[int], None]] = None,
                 row: bool = True):

        from vois.vuetify import settings

        self.value = index
        self.labels = labels
        self.tooltips = tooltips
        self.color = color if color is not None else settings.color_first
        self.on_change = on_change
        self.row = row

        self.r = []
        i = 0
        for label in self.labels:
            if self.row:
                self.r.append(v.Radio(label=label, class_="pa-0 ma-0 ml-2 mt-2 mr-6 mb-n3", color=self.color))
            else:
                self.r.append(v.Radio(label=label, class_="pa-0 ma-0 ml-2 mt-3 mr-6 mb-n2", color=self.color))
            if i < len(self.tooltips):
                self.r[i] = tooltip.tooltip(self.tooltips[i], self.r[i])
            i += 1

        super().__init__(v_model=self.value, row=self.row, class_="pa-0 ma-0", large=True,
                         color=self.color, children=self.r, style_="overflow: hidden;")

        # If requested onchange management
        if self.on_change:
            self.on_event('change', self.__internal_onchange)

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

    # Manage onchange event
    def __internal_onchange(self, widget=None, event=None, data=None):
        self.value = widget.v_model
        if self.on_change:
            self.on_change(self.value)

    # Returns the vuetify object to display (the v.RadioGroup itself)
    def draw(self):
        warnings.warn('The "draw" method is deprecated, please just use the object widget itself.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self
